#!/usr/bin/env python
"""
Convert date representations between different regional variants.

Converts between date formats called English *YYYY-MM-DD hh:mm:ss*, US-English
*MM/DD/YYYY hh:mm:ss*, French *DD/MM/YYYY hh:mm:ss*, and standard *DD.MM.YYYY
hh:mm:ss*. The routines take list of dates of different formats and return date
strings in one of the four formats.

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2015-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   date2date
   date2en
   date2fr
   date2us
   en2date
   en2fr
   en2us
   fr2date
   fr2en
   fr2us
   us2date
   us2en
   us2fr

History
    * Written Feb 2015 by Matthias Cuntz (mc (at) macu (dot) de)
    * Removed usage of date2dec and dec2date, Nov 2016, Matthias Cuntz
    * Adapted docstrings to Python 2 and 3, Nov 2016, Matthias Cuntz
    * Written fr2ascii, Mar 2018, Matthias Cuntz
    * Added us and fr keywords, renamed eng to en, Mar 2018, Matthias Cuntz
    * Added two-digit year YY, Nov 2018, Matthias Cuntz
    * Removed bug from usage of en and old name eng, Jun 2019, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * flake8, Dec 2021, Matthias Cuntz
    * Remove eng, Dec 2021, Matthias Cuntz
    * Renamed 'ascii' in function names to 'date', and to 'standard' in
      docstrings and comments, Dec 2021, Matthias Cuntz
    * New more versatile date2date function replacing ascii2ascii,
      Dec 2021, Matthias Cuntz
    * Wrapper functions between all standard formats, Dec 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz
    * Use input2array, array2input, Jun 2022, Matthias Cuntz
    * Negative years possible, Jun 2022, Matthias Cuntz
    * Fractional seconds, Jun 2022, Matthias Cuntz

"""
import time as ptime
from datetime import datetime
import numpy as np
from .helper import input2array, array2input
# from .datetime import datetime


__all__ = ['date2date',
           'date2en', 'date2fr', 'date2us',
           'en2date', 'en2fr', 'en2us',
           'fr2date', 'fr2en', 'fr2us',
           'us2date', 'us2en', 'us2fr',
           ]


def _leading_zero(l1):
    """
    Add leading 0 if necessary

    """
    return l1 if len(l1) == 2 else '0' + l1


def _ensure_year(yr, isyr2):
    """
    Ensure 4-digit year

    Years are supposed to be 4-digit years. If they have only 1 or 2 digits,
    then every year that is above the current year of the century will be taken
    as being in 1900, i.e. 90 will be taken as 1990, while all other years are
    taken in the 21st century, i.e. 20 will be 2020. 3-digit years will throw
    a ValueError.

    """
    inyr = yr.strip()
    if int(inyr) < 0:
        minus = '-'
        inyr = inyr[1:]
    else:
        minus = ''
        inyr = inyr
    lyr = len(inyr)
    if lyr == 4:
        return yr
    else:
        iyr = int(inyr)
        if iyr < 100:
            if iyr > isyr2:
                return minus + '19' + _leading_zero(inyr)
            else:
                return minus + '20' + _leading_zero(inyr)
        else:
            raise ValueError(f'3-digit years not supported: {yr}')


def date2date(edate, fr=False, format='', timesep=' ', full=False):
    """
    Convert date representations between different regional variants

    Convert date notations between standard *DD.MM.YYYY hh:mm:ss*, English
    *YYYY-MM-DD hh:mm:ss*, American *MM/DD/YYYY hh:mm:ss*, and French
    *DD/MM/YYYY hh:mm:ss* format ('%d.%m.%Y %H:%M:%S', '%Y-%m-%d %H:%M:%S',
    '%m/%d/%Y %H:%M:%S', and '%d/%m/%Y %H:%M:%S' in datetime.strftime/strptime
    format notation, respectively). The routine determines the input date
    format by the date separator '.', '-', or '/'. If the latter is detected,
    the American format is assumed. One can set *fr=True* so that '/' detects
    the French date format instead. The separator between date and time can be
    space ' ' or 'T', such as *YYYY-MM-DDThh:mm:ss*.

    Output format is given either by the calling name of the function (e.g.
    :func:`date2en`) or by the *format* keyword. It allows '', 'en', 'us', and
    'fr' for standard, English, American, and French date format, respectively.
    One can also give any format string understood by
    :func:`datetime.strftime`.

    Parameters
    ----------
    edate : array_like
        Date strings in any of three formats standard, English, or
        American/French.
    fr : bool, optional
        Input dates with '/' separators are interpreted as American format
        *MM/DD/YYYY hh:mm:ss* if False (default). Input dates with '/'
        separators are interpreted as French format *DD/MM/YYYY hh:mm:ss* if
        True.
    format : str, optional
        Output format. Can be any of '' (default), 'en', 'us', and 'fr' for
        standard *DD.MM.YYYY hh:mm:ss*, English *YYYY-MM-DD hh:mm:ss*, American
        *MM/DD/YYYY hh:mm:ss*, and French *DD/MM/YYYY hh:mm:ss* format.
        *format* can also be any format string understood by
        :func:`datetime.strftime`.
    timesep : str, optional
        Separator string between date and time if *format* is '', 'en', 'us',
        or 'fr'. Default is space ' ', but ISO 8601 uses 'T', for example.
    full : bool, optional
        Output dates are as long as input dates if False (default), e.g.
        *[YYYY-MM-DD, YYYY-MM-DD hh:mm]*. Output dates are all in full format
        *DD.MM.YYYY hh:mm:ss* if True; missing time info on input is 00 on
        output.

    Returns
    -------
    array_like
        Date representations in chosen format. Output type will be the same as
        the input type.

    Notes
    -----
    Year, month and day must be given while second, minute and hour can be
    missing (assumed zero).

    Years are supposed to be 4-digit years. If they have only 1 or 2 digits,
    then every year that is above the current year of the century will be taken
    as being in 1900, i.e. 90 will be taken as 1990, while all other years are
    taken in the 21st century, i.e. 20 will be 2020. 3-digit years will throw
    a ValueError.

    Negative years are not supported. The function uses :mod:`datetime` if a
    format string for :func:`datetime.strftime` is given. This limits the
    minimum year to the limit of :mod:`datetime` in this case, which is year
    '0001'.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(date2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990, 04.05.1786
    >>> print(", ".join(date2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00, 04.05.1786 00:00:00
    >>> print(", ".join(date2date(edate, format='en')))
    2014-11-12 12:00, 2015-03-01 17:56:00, 1990-12-01, 1786-05-04
    >>> print(", ".join(date2date(edate, format='en', full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00, 1786-05-04 00:00:00
    >>> print(", ".join(date2date(edate, format='en', full=True, timesep='T')))
    2014-11-12T12:00:00, 2015-03-01T17:56:00, 1990-12-01T00:00:00, 1786-05-04T00:00:00
    >>> print(list(date2date(edate, format='%Y%m%d%H%M%S')))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']
    >>> print(date2date(tuple(edate)))
    ('12.11.2014 12:00', '01.03.2015 17:56:00', '01.12.1990', '04.05.1786')
    >>> print(date2date(np.array(edate)))
    ['12.11.2014 12:00' '01.03.2015 17:56:00' '01.12.1990' '04.05.1786']
    >>> print(date2date(edate[0]))
    12.11.2014 12:00
    >>> print(", ".join(date2date(edate, format='us')))
    11/12/2014 12:00, 03/01/2015 17:56:00, 12/01/1990, 05/04/1786
    >>> print(", ".join(date2date(date2date(edate, format='en'),
    ...                           format='us', full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00, 05/04/1786 00:00:00
    >>> print(", ".join(date2date(edate, format='fr')))
    12/11/2014 12:00, 01/03/2015 17:56:00, 01/12/1990, 04/05/1786
    >>> print(", ".join(date2date(edate, format='fr', full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00, 04/05/1786 00:00:00

    2-digit year

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(date2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990
    >>> print(", ".join(date2date(edate, format='en')))
    2014-11-12 12:00, 2015-03-01 17:56:00, 1990-12-01
    >>> print(", ".join(date2date(edate, format='us')))
    11/12/2014 12:00, 03/01/2015 17:56:00, 12/01/1990
    >>> print(", ".join(date2date(edate, format='us', full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00
    >>> print(", ".join(date2date(edate, format='fr', full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00

    """
    # Input type and shape
    idate = input2array(edate, undef='',
                        default='01.01.0001 00:00:00')

    # Convert to given output type
    isyr2 = int(ptime.asctime()[-2:])
    odate = np.zeros_like(idate)
    for i, d in enumerate(idate):
        dd = d.strip()
        # analyse date
        if dd.count('-') > 1:
            datesep = '-'
        elif dd.count('/') > 1:
            datesep = '/'
        elif dd.count('.') > 1:
            datesep = '.'
        else:
            raise ValueError(f'No date separator could be determined: {d}')
        # analyse date time separation if time present
        if ' ' in dd:
            datetimesep = ' '
        elif 'T' in dd:
            datetimesep = 'T'
        else:
            datetimesep = ''
        # split date and time
        if datetimesep:
            ddate, dtime = dd.split(datetimesep)
        else:
            ddate = dd
            dtime = ''
        # split date
        if ddate.count(datesep) > 2:
            # negative years
            d0, d1, d2, d3 = ddate.split(datesep)
            if not d2:
                d2 = datesep + d3
            if not d1:
                raise ValueError('Unknown date format (1): '+ddate)
            if not d0:
                d0 = datesep + d1
                d1 = d2
                d2 = d3
            if not d3:
                raise ValueError('Unknown date format (2): '+ddate)
        else:
            d0, d1, d2 = ddate.split(datesep)
        if datesep == '-':
            dyear  = _ensure_year(d0, isyr2)
            dmonth = _leading_zero(d1)
            dday   = _leading_zero(d2)
        elif datesep == '/':
            if fr:
                dyear  = _ensure_year(d2, isyr2)
                dmonth = _leading_zero(d1)
                dday   = _leading_zero(d0)
            else:
                dyear  = _ensure_year(d2, isyr2)
                dmonth = _leading_zero(d0)
                dday   = _leading_zero(d1)
        elif datesep == '.':
            dyear  = _ensure_year(d2, isyr2)
            dmonth = _leading_zero(d1)
            dday   = _leading_zero(d0)
        # split time
        if dtime:
            tt = dtime.split(':')
            if len(tt) == 1:
                dhour   = _leading_zero(tt[0])
                dminute = ''
                dsecond = ''
            elif len(tt) == 2:
                dhour   = _leading_zero(tt[0])
                dminute = _leading_zero(tt[1])
                dsecond = ''
            elif len(tt) == 3:
                dhour   = _leading_zero(tt[0])
                dminute = _leading_zero(tt[1])
                if '.' in tt[2]:
                    second, microsecond = tt[2].split('.')
                    dsecond = _leading_zero(second) + '.' + microsecond
                else:
                    dsecond = _leading_zero(tt[2])
            else:
                raise ValueError(f'Only hour, minute, second supported'
                                 f' in time: {dtime}')
        else:
            # set for case that strftime is given
            dhour   = ''
            dminute = ''
            dsecond = ''
        # make output
        if format.lower() in ['', 'en', 'us', 'fr']:
            if format.lower() == 'en':
                out = dyear + '-' + dmonth + '-' + dday
            elif format.lower() == 'us':
                out = dmonth + '/' + dday + '/' + dyear
            elif format.lower() == 'fr':
                out = dday + '/' + dmonth + '/' + dyear
            else:
                out = dday + '.' + dmonth + '.' + dyear
            if dtime:
                out += timesep + dhour
                if dminute:
                    out += ':' + dminute
                else:
                    if full:
                        out += ':00'
                if dsecond:
                    out += ':' + dsecond
                else:
                    if full:
                        out += ':00'
            else:
                if full:
                    out += timesep + '00:00:00'
        else:
            if not dhour:
                dhour = '0'
            if not dminute:
                dminute = '0'
            if not dsecond:
                dsecond = '0'
                dmicrosecond = '0'
            else:
                if '.' in dsecond:
                    second, microsecond = dsecond.split('.')
                    dsecond = second
                    dmicrosecond = str(int(round(microsecond * 1000000.)))
                else:
                    dmicrosecond = '0'

            dattim = datetime(int(dyear), int(dmonth), int(dday),
                              int(dhour), int(dminute), int(dsecond),
                              int(dmicrosecond))
            # Assure 4 digit years on all platforms
            # see https://github.com/python/cpython/issues/76376
            if '%Y' in format:
                if int(dyear) < 0:
                    format04 = format.replace('%Y', '%05Y')
                else:
                    format04 = format.replace('%Y', '%04Y')
                try:
                    out = dattim.strftime(format04)
                    if ('4Y' in out) or ('5Y' in out):
                        out = dattim.strftime(format)
                except ValueError:
                    out = dattim.strftime(format)
            else:
                out = dattim.strftime(format)
        odate[i] = out

    # Return right type
    out = array2input(odate, edate, undef='')
    return out


def date2en(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with English date format output

    That means `date2date(edate, format='en', **kwargs)`;
    but *format* given in call will overwrite *format='en'*.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(date2en(edate)))
    2014-11-12 12:00, 2015-03-01 17:56:00, 1990-12-01, 1786-05-04
    >>> print(", ".join(date2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00, 1786-05-04 00:00:00
    >>> print(date2en(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(date2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'en'})
    return date2date(edate, **kwargs)


def date2fr(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with French date format output

    That means `date2date(edate, format='fr', **kwargs)`;
    but *format* given in call will overwrite *format='fr'*.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(date2fr(edate)))
    12/11/2014 12:00, 01/03/2015 17:56:00, 01/12/1990, 04/05/1786
    >>> print(", ".join(date2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00, 04/05/1786 00:00:00
    >>> print(date2fr(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(date2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'fr'})
    return date2date(edate, **kwargs)


def date2us(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with US-English date format output

    That means `date2date(edate, format='us', **kwargs)`;
    but *format* given in call will overwrite *format='us'*.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(date2us(edate)))
    11/12/2014 12:00, 03/01/2015 17:56:00, 12/01/1990, 05/04/1786
    >>> print(", ".join(date2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00, 05/04/1786 00:00:00
    >>> print(date2us(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(date2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'us'})
    return date2date(edate, **kwargs)


def en2date(edate, **kwargs):
    """
    Wrapper function for date2date with standard date format output (default)

    That means `date2date(edate, **kwargs)`.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> edate = date2date(edate, format='en')
    >>> print(", ".join(en2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990, 04.05.1786
    >>> print(", ".join(en2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00, 04.05.1786 00:00:00
    >>> print(en2date(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> edate = date2en(edate)
    >>> print(", ".join(en2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00

    """
    return date2date(edate, **kwargs)


def en2fr(edate, **kwargs):
    """
    Wrapper function for date2date with French date format output

    That means `date2date(edate, format='fr', **kwargs)`;
    but *format* given in call will overwrite *format='fr'*.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(en2fr(edate)))
    12/11/2014 12:00, 01/03/2015 17:56:00, 01/12/1990, 04/05/1786
    >>> print(", ".join(en2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00, 04/05/1786 00:00:00
    >>> print(en2fr(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(en2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'fr'})
    return date2date(edate, **kwargs)


def en2us(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with US-English date format output

    That means `date2date(edate, format='us', **kwargs)`;
    but *format* given in call will overwrite *format='us'*.

    Examples
    --------
    >>> edate = ['2014-11-12 12:00', '01.03.2015 17:56:00',
    ...          '1990-12-01', '04.05.1786']
    >>> print(", ".join(en2us(edate)))
    11/12/2014 12:00, 03/01/2015 17:56:00, 12/01/1990, 05/04/1786
    >>> print(", ".join(en2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00, 05/04/1786 00:00:00
    >>> print(en2us(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(en2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'us'})
    return date2date(edate, **kwargs)


def fr2date(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with French input and
    standard output date formats

    That means `date2date(edate, fr=True, **kwargs)`;
    but *fr* given in call will overwrite *fr=True*.

    Examples
    --------
    >>> edate = ['12/11/2014 12:00', '01/03/2015 17:56:00',
    ...          '01/12/1990', '04/05/1786']
    >>> print(", ".join(fr2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990, 04.05.1786
    >>> print(", ".join(fr2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00, 04.05.1786 00:00:00
    >>> print(fr2date(list(edate), format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']
    >>> print(fr2date(tuple(edate)))
    ('12.11.2014 12:00', '01.03.2015 17:56:00', '01.12.1990', '04.05.1786')
    >>> print(fr2date(np.array(edate)))
    ['12.11.2014 12:00' '01.03.2015 17:56:00' '01.12.1990' '04.05.1786']
    >>> print(fr2date(edate[0]))
    12.11.2014 12:00

    2-digit year

    >>> edate = ['12/11/14 12:00', '01/03/15 17:56:00', '01/12/90']
    >>> print(", ".join(fr2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990
    >>> print(", ".join(fr2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00

    """
    if 'fr' not in kwargs:
        kwargs.update({'fr': True})
    return date2date(edate, **kwargs)


def fr2en(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with French input and
    English output date formats

    That means `date2date(edate, fr=True, format='en', **kwargs)`;
    but *format* and *fr* given in call will overwrite *fr=True* and
    *format='en'*.

    Examples
    --------
    >>> edate = ['12/11/2014 12:00', '01/03/2015 17:56:00',
    ...          '01/12/1990', '04/05/1786']
    >>> print(", ".join(fr2en(edate)))
    2014-11-12 12:00, 2015-03-01 17:56:00, 1990-12-01, 1786-05-04
    >>> print(", ".join(fr2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00, 1786-05-04 00:00:00
    >>> print(fr2en(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['12/11/14 12:00', '01/03/15 17:56:00', '01/12/90']
    >>> print(", ".join(fr2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00

    """
    if 'fr' not in kwargs:
        kwargs.update({'fr': True})
    if 'format' not in kwargs:
        kwargs.update({'format': 'en'})
    return date2date(edate, **kwargs)


def fr2us(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with French input and
    US-English output date formats

    That means `date2date(edate, fr=True, format='us', **kwargs)`;
    but *format* and *fr* given in call will overwrite *fr=True* and
    *format='us'*.

    Examples
    --------
    >>> edate = ['12/11/2014 12:00', '01/03/2015 17:56:00',
    ...          '01/12/1990', '04/05/1786']
    >>> print(", ".join(fr2us(edate)))
    11/12/2014 12:00, 03/01/2015 17:56:00, 12/01/1990, 05/04/1786
    >>> print(", ".join(fr2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00, 05/04/1786 00:00:00
    >>> print(fr2us(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['12/11/14 12:00', '01/03/15 17:56:00', '01/12/90']
    >>> print(", ".join(fr2us(edate, full=True)))
    11/12/2014 12:00:00, 03/01/2015 17:56:00, 12/01/1990 00:00:00

    """
    if 'fr' not in kwargs:
        kwargs.update({'fr': True})
    if 'format' not in kwargs:
        kwargs.update({'format': 'us'})
    return date2date(edate, **kwargs)


def us2date(edate, **kwargs):
    """
    Wrapper function for date2date with standard date format output (default)

    That means `date2date(edate, **kwargs)`.

    Examples
    --------
    >>> edate = ['11/12/2014 12:00', '01.03.2015 17:56:00',
    ...          '12/01/1990', '1786-05-04']
    >>> print(", ".join(us2date(edate)))
    12.11.2014 12:00, 01.03.2015 17:56:00, 01.12.1990, 04.05.1786
    >>> print(", ".join(us2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00, 04.05.1786 00:00:00
    >>> print(us2date(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> edate = date2en(edate)
    >>> print(", ".join(us2date(edate, full=True)))
    12.11.2014 12:00:00, 01.03.2015 17:56:00, 01.12.1990 00:00:00

    """
    return date2date(edate, **kwargs)


def us2en(edate, **kwargs):
    """
    Wrapper function for :func:`date2date` with English date format output

    That means `date2date(edate, format='en', **kwargs)`;
    but *format* given in call will overwrite *format='en'*.

    Examples
    --------
    >>> edate = ['11/12/2014 12:00', '01.03.2015 17:56:00',
    ...          '12/01/1990', '1786-05-04']
    >>> print(", ".join(us2en(edate)))
    2014-11-12 12:00, 2015-03-01 17:56:00, 1990-12-01, 1786-05-04
    >>> print(", ".join(us2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00, 1786-05-04 00:00:00
    >>> print(us2en(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(us2en(edate, full=True)))
    2014-11-12 12:00:00, 2015-03-01 17:56:00, 1990-12-01 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'en'})
    return date2date(edate, **kwargs)


def us2fr(edate, **kwargs):
    """
    Wrapper function for date2date with French date format output

    That means `date2date(edate, format='fr', **kwargs)`;
    but *format* given in call will overwrite *format='fr'*.

    Examples
    --------
    >>> edate = ['11/12/2014 12:00', '01.03.2015 17:56:00',
    ...          '12/01/1990', '1786-05-04']
    >>> print(", ".join(us2fr(edate)))
    12/11/2014 12:00, 01/03/2015 17:56:00, 01/12/1990, 04/05/1786
    >>> print(", ".join(us2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00, 04/05/1786 00:00:00
    >>> print(us2fr(edate, format='%Y%m%d%H%M%S'))
    ['20141112120000', '20150301175600', '19901201000000', '17860504000000']

    >>> edate = ['14-11-12 12:00', '01.03.15 17:56:00', '90-12-01']
    >>> print(", ".join(us2fr(edate, full=True)))
    12/11/2014 12:00:00, 01/03/2015 17:56:00, 01/12/1990 00:00:00

    """
    if 'format' not in kwargs:
        kwargs.update({'format': 'fr'})
    return date2date(edate, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec import date2dec
from pyjams.const import mmol_co2, mmol_h2o, mmol_air, cheat_air, latentheat_vaporization, T0
from scipy.interpolate import splrep, splint
from pyjams.air_humidity import esat


def profile2storage(fluxfile, fluxfile2, profilefile, outdir, heights, CO2=None,
                    H2O=None, T=None, rH=None, delimiter=[',',',',','],
                    skiprows=[1,1,1], format=['ascii','ascii','ascii'],
                    undef=-9999, plot=False):
    '''
    Calculates storage fluxes for changes in CO2, H2O, air temperature and air
    moisture from profile data or meteorological data to correct Eddy
    Covariance fluxes. FLux files from EddySoft and from fluxflag are needed as
    well as a file with the profile or meteo data. Fluxes will be updated with
    the respective storage fluxes and saved in a new file. Multiple application
    of this routine with different profile or meteo files are possible to
    correct e.g. the CO2, H2O and latent heat fluxes with profile data of CO2
    and H2O concentrations and afterwards the H flux with temperature data from
    another file.


    Definition
    ----------
    profile2storage(fluxfile, fluxfile2, profilefile, outdir, heights, CO2=None,
                    H2O=None, T=None, rH=None, delimiter=[',',',',','],
                    skiprows=[1,1,1], format=['ascii','ascii','ascii'],
                    undef=-9999, plot=False):


    Input
    -----
    fluxfile    str, path and file name of fluxflag output file containing
                fluxes and flags. These fluxes will be updated by the storage
                fluxes and saved as a new file
    fluxfile2   str, path and file name of EddyFlux output file (timestep
                checked) containing original fluxes
    profilefile str, path and file name of the profile file or meteorology file
                containing CO2, H2O, T or rH values to compute the profile
                storage from
    outdir      str, path of the output folder
    heights     list of floats, observation heights of the profile [m],
                increasing e.g. [0.5,1.0,10.0,20.0].
    CO2         list of int, column numbers of CO2 concentrations for the
                different heights (in the same order) [mumol/mol] in profilefile,
                column number starts with 0 which is first data column.
    H2O         list of int, column numbers of H2O concentrations for the
                different heights (in the same order) [mmol/mol] in profilefile,
                column number starts with 0 which is first data column.
    T           list of int, column numbers of air temperatures for the
                different heights (in the same order) [degC] in profilefile,
                column number starts with 0 which is first data column.
    rH          list of int, column numbers of relative humidity for the
                different heights (in the same order) [%] in profilefile,
                column number starts with 0 which is first data column. The
                calculation of air vapour energy storage change within the
                profile works only when T is given as well.


    Optional Input
    --------------
    delimiter   list of str, delimiters of fluxfile, fluxfile and profilefile
                (default: [',',',',','])
    skiprows    list of int, lines to skip at the beginning of fluxfile,
                fluxfile and profilefile, e.g. header lines (default: [1,1,1])
    format      list of str, time formats of fluxfile, fluxfile and profilefile,
                'ascii' and 'eng' possible (default: ['ascii','ascii','ascii'])
    undef       int/float, missing value of fluxfile, fluxfile and profilefile
                (default: -9999, np.nan is not possible)
    plot        bool, if True performs plotting (default: False)


    Output
    ------
    flux+stor.csv file containing fluxes and flags where storage fluxes are
                  added in an additional column and storage fluxes are appended
                  to the end of the file


    Restrictions
    ------------
    Works only with half hourly time steps, all files in sync


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Sep 2014
    '''
    ###########################################################################
    # time interval
    int = 30.
    dt  = int*60.

    if plot:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf

    ###########################################################################
    # reading input files
    # fluxes to correct for storage changes
    d1 = np.loadtxt(fluxfile, dtype='|S100', delimiter=delimiter[0])
    # original flux file from EddyFlux containing air density rho_a
    d2 = np.loadtxt(fluxfile2, dtype='|S100', delimiter=delimiter[1])
    # file containing profile data (can be meteo file if no profile available)
    d3 = np.loadtxt(profilefile, dtype='|S100', delimiter=delimiter[2])

    assert (d1.shape[1]==11) | (d1.shape[1]==19), 'profile2storage: fluxfile must be from fluxflag or profiletostorage and have 11 or 19 cols'
    assert d2.shape[1]==68, 'profile2storage: fluxfile2 must be from EddyFlux and have 68 cols'
    assert d1.shape[0]==d2.shape[0], 'profile2storage: fluxfile and fluxfile2 must be in sync'
    assert d1.shape[0]==d3.shape[0], 'profile2storage: fluxfile and profilefile must be in sync'
    assert (((H2O==None) & (rH==None)) ^ ((H2O!=None) ^ (rH!=None))), 'profile2storage: give either H2O or rH, both would be double correction'

    if format[0]=='ascii':
        datev   = date2dec(ascii=d1[skiprows[0]:,0])
    elif format[0]=='eng':
        datev   = date2dec(eng=d1[skiprows[0]:,0])
    else:
        raise ValueError('profile2storage: unknown format')
    if format[2]=='ascii':
        datem   = date2dec(ascii=d2[skiprows[2]:,0])
    elif format[2]=='eng':
        datem   = date2dec(eng=d2[skiprows[2]:,0])
    else:
        raise ValueError('profile2storage: unknown format')

    flux1 = np.where(d1[skiprows[0]:,1:]=='', str(undef), d1[skiprows[0]:,1:]).astype(np.float)
    flux2 = np.where(d2[skiprows[1]:,1:]=='', str(undef), d2[skiprows[1]:,1:]).astype(np.float)
    prof  = np.where(d3[skiprows[2]:,1:]=='', str(undef), d3[skiprows[2]:,1:]).astype(np.float)

    flux1 = np.ma.array(flux1, mask=flux1==undef, hard_mask=True)
    flux2 = np.ma.array(flux2, mask=flux2==undef)
    prof  = np.ma.array(prof,  mask=prof==undef)

    ###########################################################################
    # assign variables
    if d1.shape[1]==11:
        H, Hflag   = flux1[:,0], flux1[:,1]
        Le, Leflag = flux1[:,2], flux1[:,3]
        E, Eflag   = flux1[:,4], flux1[:,5]
        C, Cflag   = flux1[:,6], flux1[:,7]
    else:
        H, Hflag   = flux1[:,0], flux1[:,2]
        Le, Leflag = flux1[:,3], flux1[:,5]
        E, Eflag   = flux1[:,6], flux1[:,8]
        C, Cflag   = flux1[:,9], flux1[:,11]
    p          = flux2[:,58] # [hPa]
    rho        = flux2[:,62] # [kg/m3]

    ###########################################################################
    # prepare output array
    d4 = np.copy(d1)
    if d1.shape[1]==11:
        temp = np.empty((d1.shape[0],4), dtype='|S100')
        temp[:] = ' '*(11-len(str(undef)))+str(undef)
        temp[0,:] = ['       H+sT','     LE+sLE','       E+sE','       C+sC']
        d4 = np.insert(d4, [2,4,6,8], temp, axis=1)

        temp[0,:] = ['         sT','        sLE','         sE','         sC']
        d4 = np.append(d4, temp, axis=1)

    ###########################################################################
    # calls
    if CO2:
        CO2 = prof[:,CO2]
        assert CO2.shape[1]==len(heights), 'profile2storage: number of CO2 cols must equal heights'
        # calculate storage flux and storage flux flag
        sfCO2     = stor2flux(CO2, rho, heights, dt, 'CO2')
        sfCO2flag = sfCO2.mask.astype(np.int)
        # add to eddy flux
        newC      = C + np.ma.filled(sfCO2, 0)
        # format and write into output array
        newC_str  = np.array(['%11.5f'%x for x in np.ma.filled(newC, undef)])
        newC_str  = np.where(newC_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newC_str)
        sfCO2_str = np.array(['%11.5f'%x for x in np.ma.filled(sfCO2, undef)])
        sfCO2_str = np.where(sfCO2_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfCO2_str)
        d4[skiprows[0]:,11] = newC_str
        d4[skiprows[0]:,18] = sfCO2_str

        if plot:
            storplot(CO2, datev, heights, C, sfCO2, newC, 'storageCO2.pdf', pdf, plt, mpl, outdir)

    if H2O:
        H2O = prof[:,H2O]
        assert H2O.shape[1]==len(heights), 'profile2storage: number of H2O cols must equal heights'
        # calculate storage flux and storage flux flag
        sfH2O     = stor2flux(H2O, rho, heights, dt, 'H2O')
        sfH2O_Wm2 = sfH2O * mmol_h2o * latentheat_vaporization /1.e6
        sfH2Oflag = sfH2O.mask.astype(np.int)
        # add to eddy flux
        newE      = E + np.ma.filled(sfH2O, 0)
        newLe     = Le + np.ma.filled(sfH2O_Wm2, 0)
        # format and write into output array
        newE_str      = np.array(['%11.5f'%x for x in np.ma.filled(newE, undef)])
        newLe_str     = np.array(['%11.5f'%x for x in np.ma.filled(newLe, undef)])
        sfH2O_str     = np.array(['%11.5f'%x for x in np.ma.filled(sfH2O, undef)])
        sfH2O_Wm2_str = np.array(['%11.5f'%x for x in np.ma.filled(sfH2O_Wm2, undef)])
        newE_str      = np.where(newE_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newE_str)
        newLe_str     = np.where(newLe_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newLe_str)
        sfH2O_str     = np.where(sfH2O_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfH2O_str)
        sfH2O_Wm2_str = np.where(sfH2O_Wm2_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfH2O_Wm2_str)
        d4[skiprows[0]:,8]  = newE_str
        d4[skiprows[0]:,17] = sfH2O_str
        d4[skiprows[0]:,5]  = newLe_str
        d4[skiprows[0]:,16] = sfH2O_Wm2_str

        if plot:
            storplot(H2O, datev, heights, E, sfH2O, newE, 'storageH2O.pdf', pdf, plt, mpl, outdir)

    if T:
        T   = prof[:,T]
        assert T.shape[1]==len(heights), 'profile2storage: number of T cols must equal heights'
        # calculate storage flux and storage flux flag
        sfT       = stor2flux(T, rho, heights, dt, 'T')
        sfTflag   = sfT.mask.astype(np.int)
        # add to eddy flux
        newH      = H + np.ma.filled(sfT, 0)
        # format and write into output array
        newH_str  = np.array(['%11.5f'%x for x in np.ma.filled(newH, undef)])
        newH_str  = np.where(newH_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newH_str)
        sfT_str   = np.array(['%11.5f'%x for x in np.ma.filled(sfT, undef)])
        sfT_str   = np.where(sfT_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfT_str)
        d4[skiprows[0]:,2]  = newH_str
        d4[skiprows[0]:,15] = sfT_str

        if plot:
            storplot(T, datev, heights, H, sfT, newH, 'storageT.pdf', pdf, plt, mpl, outdir)

    if rH:
        rH  = prof[:,rH]
        assert rH.shape[1]==len(heights), 'profile2storage: number of rH cols must equal heights'
        # calculate specific humidity
        vapourpressure = esat(T+T0)*(rH/100.)/100. #[hPa]
        specifichumidity = (mmol_h2o/mmol_air*vapourpressure) / (p-(1.-mmol_h2o/mmol_air)*vapourpressure)
        # calculate storage flux and storage flux flag
        sfrH_Wm2  = stor2flux(specifichumidity, rho, heights, dt, 'rH')
        sfrH      = sfrH_Wm2 * 1.e6 / (mmol_h2o * latentheat_vaporization)
        sfrHflag  = sfrH.mask.astype(np.int)
        # add to eddy flux
        newE      = E + np.ma.filled(sfrH, 0)
        newLe     = Le + np.ma.filled(sfrH_Wm2, 0)
        # format and write into output array
        newE_str     = np.array(['%11.5f'%x for x in np.ma.filled(newE, undef)])
        newLe_str    = np.array(['%11.5f'%x for x in np.ma.filled(newLe, undef)])
        sfrH_str     = np.array(['%11.5f'%x for x in np.ma.filled(sfrH, undef)])
        sfrH_Wm2_str = np.array(['%11.5f'%x for x in np.ma.filled(sfrH_Wm2, undef)])
        newE_str     = np.where(newE_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newE_str)
        newLe_str    = np.where(newLe_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), newLe_str)
        sfrH_str     = np.where(sfrH_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfrH_str)
        sfrH_Wm2_str = np.where(sfrH_Wm2_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), sfrH_Wm2_str)
        d4[skiprows[0]:,8]  = newE_str
        d4[skiprows[0]:,17] = sfrH_str
        d4[skiprows[0]:,5]  = newLe_str
        d4[skiprows[0]:,16] = sfrH_Wm2_str

        if plot:
            storplot(rH, datev, heights, E, sfH2O, newE, 'storagerH.pdf', pdf, plt, mpl, outdir)

    ###########################################################################
    # write output
    np.savetxt('%s/flux+stor.csv'%outdir, d4, '%s', delimiter=',')

def stor2flux(concentrations, rho, heights, dt, constituent='CO2'):
    '''
    '''
    xb = 0.0                # bottom height of interpolation
    xe = np.amax(heights)   # top height of interpolation

    if constituent=='CO2':
        # mole volume [m3/mol] = mmol_co2[g/mol]/(rho[kg/m3]*1000.)
        m = mmol_co2/(rho*1000.)
    elif constituent=='H2O':
        # mole volume [m3/mol] = mmol_h2o[g/mol]/(rho[kg/m3]*1000.)
        m = mmol_h2o/(rho*1000.)
    elif constituent=='T':
        # 1/energy content of the air [1/(J/m3 K)] = 1/ (rho[kg/m3]*heat capacity of air [J/kg K])
        m = 1./(rho*cheat_air)
    elif constituent=='rH':
        # 1/energy content of vapor [1/(J/m3)] = 1/ (rho[kg/m3] * specific heat of vaporization of water [J/kg])
        m = 1./(rho * latentheat_vaporization)
    else:
        raise ValueError('stor2flux: unknown constituent')

    ###########################################################################
    # calculate storage for every time step
    storage, sf = np.ma.masked_all_like(rho), np.ma.masked_all_like(rho)
    for i,item in enumerate(concentrations):
        if not item.mask.any():
            # if only one height given, take box approach (splrep does not work)
            if len(heights)==1:
                storage[i] = item*heights
            # else interpolate nicely :-)
            else:
                tck        = splrep(heights,item,xb=xb,xe=xe,k=1)
                storage[i] = splint(xb,xe,tck)

    ###########################################################################
    # calculate storage flux
    # storage flux per time step
    # for CO2: [mumol/m*2] = [mumol/mol*m]/[m3/mol]
    # for H2O: [mmol/m*2] = [mmol/mol*m]/[m3/mol]
    # for T:   [J/m*2] = [K*m]/[1/(J/m3 K)]
    # for rH:  [J/m*2] = [m]/[1/(J/m3)]
    sf[1:] = storage[:-1]/m[:-1] - storage[1:]/m[1:]
    sf[0]  = sf[1]
    # storage flux per second
    # for CO2: [mumol/(m2*s)]
    # for H2O: [mmol/(m2*s)]
    # for T:   [J/(m2*s)]=[W/m*2]
    # for rH:  [J/(m2*s)]=[W/m*2]
    sf     = sf/dt

    return sf

def storplot(conc, date, heights, oriflux, storflux, newflux, name, pdf, plt, mpl, outdir):
    '''
    '''
    majticks = mpl.dates.MonthLocator(bymonthday=1)
    format_str='%d %m %Y %H:%M'
    date01 = date2dec(yr=1, mo=1, dy=2, hr=0, mi=0, sc=0)

    conc = np.ma.copy(conc.transpose())
    date = np.ma.copy(date-date01)

    pp1 = pdf.PdfPages(outdir+'/'+name)
    fig1 = plt.figure(name)
    sub1 = fig1.add_subplot(211)
    for i, item in enumerate(conc):
        sub1.plot(date, item, label='%2.1f m'%(heights[i]))
    plt.legend(loc='best')

    sub2 = fig1.add_subplot(212)
    sub2.axhline(y=0, xmin=0, xmax=1, color='k')
    sub2.plot(date, oriflux, 'b-', label='original')
    sub2.plot(date, storflux, 'r-', label='storage')
    sub2.plot(date, newflux, 'g-', label='new')
    plt.legend(loc='best')

    sub1.set_xlim(date[0],date[-1])
    sub1.xaxis.set_major_locator(majticks)
    sub1.xaxis.set_major_formatter(mpl.dates.DateFormatter(format_str))
    sub2.set_xlim(date[0],date[-1])
    sub2.xaxis.set_major_locator(majticks)
    sub2.xaxis.set_major_formatter(mpl.dates.DateFormatter(format_str))
    fig1.autofmt_xdate()

    plt.show()
    fig1.savefig(pp1, format='pdf')
    pp1.close()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
import numpy as np
from pyjams.fsread import sread
from pyjams.fsread import fread
import csv
import os as os
from math import pi
from scipy.stats import skew
import sys


def planarfit(indirpf, rawfile, outfile, pfmat='pfitmatrix.csv',
              pf0file='pfitdata0.csv', pf1file='pfitdata1.csv',
              pf2file='pfitdata2.csv', histsteps=50, plot=False):
    '''
    Extracts raw wind speeds from the raw flux file of EddyFlux as input for
    EDDYPFit. When EDDYFit is finished, the script loads the results and
    do plots. If user is satisfied, results are saved.


    Definition
    ----------
    planarfit(indirpf, rawfile, outfile, pfmat='pfitmatrix.csv',
              pf0file='pfitdata0.csv', pf1file='pfitdata1.csv',
              pf2file='pfitdata2.csv', histsteps=50):

    Input
    -----
    indirpf     str, path of the folder where results will be saved
    rawfile     str, path of the file with raw wind speeds from EddyFlux
    outfile     str, name of the output file


    Optional Input
    --------------
    pfmat       str, name of the pfitmatix file, default: 'pfitmatrix.csv'
    pf0file     str, name of the original wind speed file of EDDYPFit, default: 'pfitdata0.csv'
    pf1file     str, name of the one plane fit wind speed file of EDDYPFit, default: 'pfitdata1.csv'
    pf2file     str, name of the sectorial fit wind speed file of EDDYPFit, default: 'pfitdata2.csv'
    histstep    int, histogram steps for plotting (default=50)


    Output
    ------
    X_pfit.pdf  plot with planar fit
    X_uvw.csv   file with raw wind speeds
    X_wd.pdf    plot with wind rose
    X_wdis.pdf  plot with wind speed distributions


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''
    ############################################################################
    # reading raw file
    uvw   = np.array(fread('%s' %rawfile, skip=1, cskip=13, nc=3))
    wdhor = np.array(fread('%s' %rawfile, skip=1, cskip=20, nc=1))
    header = np.array(sread('%s' %rawfile, cskip=13, nc=3), dtype='|S5')

    ############################################################################
    # coordinate transformation for u+v
    alpha = 120.*np.pi/180. #rotation angle in rad
    uvw_trans = np.copy(uvw)
    uvw_trans[:,0] = np.where((uvw[:,0]!=-9999) & (uvw[:,1]!=-9999),
                              uvw[:,0]*np.cos(alpha)+uvw[:,1]*np.sin(alpha),
                              -9999)# u
    uvw_trans[:,1] = np.where((uvw[:,0]!=-9999) & (uvw[:,1]!=-9999),
                              -uvw[:,0]*np.sin(alpha)+uvw[:,1]*np.cos(alpha),
                              -9999)# v

    ############################################################################
    # writing uvw file with wind speed for EDDYPFit
    file1 = open('%s/%s_uvw.csv' %(indirpf,outfile[:-4]), 'w')
    output = csv.writer(file1)
    output.writerow(header[0])
    for i in xrange(np.shape(uvw_trans)[0]):
        output.writerow(uvw_trans[i])
    file1.close()

    ############################################################################
    # user input to continue
    print("Do EddyPFit with the 'uvw.csv' file now!")
    ui1 = raw_input("Ready or quit (y/n)?: ").lower()
    if ui1 != "y":
        sys.exit()

    ############################################################################
    # reading pfit files
    header0 = np.array(sread('%s/%s' %(indirpf,pf0file)) , dtype='|S1')
    uvw0    = np.array(fread('%s/%s' %(indirpf,pf0file), skip=1), dtype=np.float)
    uvw0_trans = np.copy(uvw0)
    uvw0_trans[:,0] = uvw0[:,0]*np.cos(alpha) - uvw0[:,1]*np.sin(alpha)
    uvw0_trans[:,1] = uvw0[:,0]*np.sin(alpha) + uvw0[:,1]*np.cos(alpha)

    header1 = np.array(sread('%s/%s' %(indirpf,pf1file)) , dtype='|S1')
    uvw1    = np.array(fread('%s/%s' %(indirpf,pf1file), skip=1), dtype=np.float)
    uvw1_trans = np.copy(uvw1)
    uvw1_trans[:,0] = uvw1[:,0]*np.cos(alpha) - uvw1[:,1]*np.sin(alpha)
    uvw1_trans[:,1] = uvw1[:,0]*np.sin(alpha) + uvw1[:,1]*np.cos(alpha)

    header2 = np.array(sread('%s/%s' %(indirpf,pf2file)) , dtype='|S1')
    uvw2    = np.array(fread('%s/%s' %(indirpf,pf2file), skip=1), dtype=np.float)
    uvw2_trans = np.copy(uvw2)
    uvw2_trans[:,0] = uvw2[:,0]*np.cos(alpha) - uvw2[:,1]*np.sin(alpha)
    uvw2_trans[:,1] = uvw2[:,0]*np.sin(alpha) + uvw2[:,1]*np.cos(alpha)

    ############################################################################
    # plots
    # define grid
    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec
        from matplotlib.mlab import griddata
        import matplotlib.cm as cm
        import matplotlib.mlab as mlab
        import matplotlib.backends.backend_pdf as pdf
        x0 = np.linspace(np.min(np.minimum(uvw0_trans[:,0],uvw0_trans[:,1])),
                         np.max(np.maximum(uvw0_trans[:,0],uvw0_trans[:,1])),500)
        y0 = np.linspace(np.min(np.minimum(uvw0_trans[:,0],uvw0_trans[:,1])),
                         np.max(np.maximum(uvw0_trans[:,0],uvw0_trans[:,1])),500)

        x1 = np.linspace(np.min(np.minimum(uvw1_trans[:,0],uvw1_trans[:,1])),
                         np.max(np.maximum(uvw1_trans[:,0],uvw1_trans[:,1])),500)
        y1 = np.linspace(np.min(np.minimum(uvw1_trans[:,0],uvw1_trans[:,1])),
                         np.max(np.maximum(uvw1_trans[:,0],uvw1_trans[:,1])),500)

        x2 = np.linspace(np.min(np.minimum(uvw2_trans[:,0],uvw2_trans[:,1])),
                         np.max(np.maximum(uvw2_trans[:,0],uvw2_trans[:,1])),500)
        y2 = np.linspace(np.min(np.minimum(uvw2_trans[:,0],uvw2_trans[:,1])),
                         np.max(np.maximum(uvw2_trans[:,0],uvw2_trans[:,1])),500)

        # grid the data.
        z0 = griddata(uvw0_trans[:,0],uvw0_trans[:,1],uvw0_trans[:,2],x0,y0, interp='linear')
        z1 = griddata(uvw1_trans[:,0],uvw1_trans[:,1],uvw1_trans[:,2],x1,y1, interp='linear')
        z2 = griddata(uvw2_trans[:,0],uvw2_trans[:,1],uvw2_trans[:,2],x2,y2, interp='linear')

        # plotting contours
        fig1 = plt.figure(1, figsize=(6,13))
        sub1 = fig1.add_subplot(311, aspect=1)
        fillings = sub1.contourf(x0,y0,z0,20,cmap=plt.cm.jet)
        scat = sub1.scatter(uvw0_trans[:,0],uvw0_trans[:,1],marker='o',c='b',s=0.2,zorder=10)
        cbar = fig1.colorbar(fillings, orientation='vertical')
        xlimits = sub1.get_xlim()
        sub1.plot(np.array([xlimits[0],xlimits[1]]),np.array([0,0]), c='k')
        ylimits = sub1.get_ylim()
        sub1.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='k')
        sub1.set_title('Original wind components\nwith point data')
        cbar.set_label('w [m/s]')
        plt.ylabel('v [m/s]')

        sub2 = fig1.add_subplot(312, aspect=1)
        fillings = sub2.contourf(x1,y1,z1,20,cmap=plt.cm.jet)
        cbar = fig1.colorbar(fillings, orientation='vertical')
        xlimits = sub2.get_xlim()
        sub2.plot(np.array([xlimits[0],xlimits[1]]),np.array([0,0]), c='k')
        ylimits = sub2.get_ylim()
        sub2.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='k')
        sub2.set_title('One plane')
        cbar.set_label('w [m/s]')
        plt.ylabel('v [m/s]')

        sub3 = fig1.add_subplot(313, aspect=1)
        fillings = sub3.contourf(x2,y2,z2,20,cmap=plt.cm.jet)
        cbar = fig1.colorbar(fillings, orientation='vertical')
        xlimits = sub3.get_xlim()
        sub3.plot(np.array([xlimits[0],xlimits[1]]),np.array([0,0]), c='k')
        ylimits = sub3.get_ylim()
        sub3.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='k')
        sub3.set_title('Sectorial')
        cbar.set_label('w [m/s]')
        plt.xlabel('u [m/s]')
        plt.ylabel('v [m/s]')

        # plotting histograms
        mi = np.min(np.array([np.min(uvw0_trans[:,2]),np.min(uvw1_trans[:,2]),
                              np.min(uvw2_trans[:,2])]))
        ma = np.max(np.array([np.max(uvw0_trans[:,2]),np.max(uvw1_trans[:,2]),
                              np.max(uvw2_trans[:,2])]))
        steps = np.abs((ma-mi)/histsteps)
        bins = np.arange(mi,ma+steps,steps)

        fig2 = plt.figure(2, figsize=(6,13))
        fig2.subplots_adjust(hspace=0.3)
        sub4 = fig2.add_subplot(311)
        n0, bins0, patches0 = sub4.hist(uvw0_trans[:,2], bins, color= 'b', histtype='bar')
        ylimits = sub4.get_ylim()
        sub4.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='y', lw=3)
        plt.ylabel('count')
        plt.title('Original w-component:\navg(w)= %.2f, var(w)= %.4f, skew(w)= %.4f'
                  %(np.mean(uvw0_trans[:,2]), np.var(uvw0_trans[:,2]), skew(uvw0_trans[:,2])))

        sub5 = fig2.add_subplot(312)
        n1, bins1, patches1 = sub5.hist(uvw1_trans[:,2], bins, color= 'g', histtype='bar')
        ylimits = sub5.get_ylim()
        sub5.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='y', lw=3)
        plt.ylabel('count')
        plt.title('One plane w-component:\navg(w)= %.2f, var(w)= %.4f, skew(w)= %.4f'
                  %(np.mean(uvw1_trans[:,2]), np.var(uvw1_trans[:,2]), skew(uvw1_trans[:,2])))

        sub6 = fig2.add_subplot(313)
        n2, bins2, patches2 = sub6.hist(uvw2_trans[:,2], bins, color= 'r', histtype='bar')
        ylimits = sub6.get_ylim()
        sub6.plot(np.array([0,0]),np.array([ylimits[0],ylimits[1]]), c='y', lw=3)
        plt.xlabel('Classes [m/s]')
        plt.ylabel('count')
        plt.title('Sectorial w-component:\navg(w)= %.2f, var(w)= %.4f, skew(w)= %.4f'
                  %(np.mean(uvw2_trans[:,2]), np.var(uvw2_trans[:,2]), skew(uvw2_trans[:,2])))

        # wind rose
        fig3 = plt.figure(3, figsize=(6,6))
        pol = fig3.add_subplot(111, polar=True)
        hist, bin_edges= np.histogram(wdhor, bins=36, range=(0,360))
        x = 90-np.arange(5,365,10)
        x = [i*pi/180. for i in x]  # convert to radians
        pol.bar(x, hist, width=10*pi/180)
        pol.set_xticklabels([r'$\sf{90\degree}$',r'$\sf{45\degree}$',r'$\sf{0\degree}$',
                             r'$\sf{315\degree}$',r'$\sf{270\degree}$',r'$\sf{225\degree}$',
                             r'$\sf{180\degree}$',r'$\sf{135\degree}$'], fontsize=15)
        plt.title('Horizontal wind direction frequency')

        plt.show()

    ############################################################################
    # user input for saving results
    print("Satisfied with the fit?\ny will save the figures, n will exit without saving!")
    ui2 = raw_input("(y/n)?: ").lower()
    if ui2 != "y":
        sys.exit()

    ############################################################################
    # save results
    if plot:
        pp1 = pdf.PdfPages('%s/%s_pfit.pdf'%(indirpf,outfile[:-4]))
        pp2 = pdf.PdfPages('%s/%s_wdis.pdf'%(indirpf,outfile[:-4]))
        pp3 = pdf.PdfPages('%s/%s_wd.pdf'%(indirpf,outfile[:-4]))
        fig1.savefig(pp1, format='pdf')
        fig2.savefig(pp2, format='pdf')
        fig3.savefig(pp3, format='pdf')
        pp1.close()
        pp2.close()
        pp3.close()

    print("Rename EddyPFit files?")
    ui3 = raw_input("(y/n)?: ").lower()
    if ui3 != "y":
        sys.exit()

    os.rename('%s/%s' %(indirpf,pf0file), '%s/%s_%s.csv' %(indirpf, outfile[:-4], pf0file[:-4]))
    os.rename('%s/%s' %(indirpf,pf1file), '%s/%s_%s.csv' %(indirpf, outfile[:-4], pf1file[:-4]))
    os.rename('%s/%s' %(indirpf,pf2file), '%s/%s_%s.csv' %(indirpf, outfile[:-4], pf2file[:-4]))
    os.rename('%s/%s' %(indirpf,pfmat), '%s/%s_%s.csv' %(indirpf, outfile[:-4], pfmat[:-4]))

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT License.

    Copyright (c) 2016 Juliane Mai, Dmitri Kavetski, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
import numpy as np


__all__ = ['sample_ep',     # sample from exponential power distribution of Box and Tiao (1992)
           'sample_sep',    # sample from skew exponential power distribution
           'sample_sep_fs', # sample from skew exponential power distribution after Fernandez and Steel (1998)
           'sample_st',     # sample from the skewed Student-t distribution
           'sample_st_fs',  # sample from skewed Student-t distribution after Fernandez and Steel (1998)
           'sample_t']      # sample from Student-t distribution


# --------------------------------------------------------------------


def sample_ep(nn, loc=0., sca=1., beta=0., sig=None):
    """
        Samples from the exponential power function of Box and Tiao (1992) ep(loc,sca,beta) are drawn.

        The distribution is symmetric.

        The kurtosis of this sample can be calculated by
             bb   = 2./(beta+1.)
             kurt = (sp.gamma(5./bb)*sp.gamma(1./bb))/((sp.gamma(3./bb))**2) - 3.


        Definition
        ----------
        def sample_ep(nn, loc=0., sca=1., beta=0., sig=None):


        Input
        -----
        nn         number of samples


        Optional Input
        --------------
        loc        location
        sca        scale
        beta       parameter controlling kurtosis
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples drawn from exponential power function


        Examples
        --------


        History
        -------
        Written  JM, May 2016
    """

    if sig is not None: sca = sig

    EP01 = sample_ep01(nn, beta=beta)
    EP   = loc + sca * EP01

    return EP


def sample_ep01(nn, beta=0.):
    """
        Samples from the exponential power function of Box and Tiao (1992) ep(0,1,beta) are drawn.

        The distribution is symmetric.

        The kurtosis of this sample can be calculated by
             bb   = 2./(beta+1.)
             kurt = (sp.gamma(5./bb)*sp.gamma(1./bb))/((sp.gamma(3./bb))**2) - 3.


        Definition
        ----------
        def sample_ep01(nn, beta=0.):


        Input
        -----
        nn         number of samples


        Optional Input
        --------------
        beta       parameter which controls the kurtosis


        Output
        ------
        Samples drawn from exponential power function


        Examples
        --------


        Literature
        --------
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.
        --> Steps (1)-(3) described on page 5


        History
        -------
        Written  JM, May 2016
    """
    import scipy.special as sp

    bb = (1.+beta)/2.

    # (1) Generate a sample gg from gamma distribution with shape
    #     parameter (1+beta)/2 and scale parameter 1.
    gg = np.random.gamma(bb, 1.0, nn)

    # (2) Generate a random sign ss (+1 or -1) with equal probability.
    ss = np.where(np.random.rand(nn) < 0.5, -1., 1.)

    # (3) Compute EP which is a sample from the exponential power distribution EP(0,1,beta)
    EP01 = ss * np.abs(gg)**bb * np.sqrt(sp.gamma(bb)) / np.sqrt(sp.gamma(3.*bb))

    return EP01


# --------------------------------------------------------------------


def sample_sep(nn, loc=0., sca=1., xi=1., beta=0., sig=None):
    """
        Samples from the skew exponential power distribution with
        given location, scale, and parameters controlling skewness and kurtosis.

        The xi parameter needs to be positive.
             1 = symmetric
            >1 = right skewed
            <1 = left skewed

        The beta parameter needs to be from [-1,1]
             1 = fat tail
             0 = medium tails
            -1 = thin tail

        If xi = 1 and beta = 0, the SEP is the Gaussian distribution.


        Definition
        ----------
        def sample_sep(nn, loc=0., sca=1., xi=1., beta=0., sig=None):


        Input
        -----
        nn         number of samples


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         parameter which controls the skewness
        beta       parameter which controls the kurtosis
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples from the (general) skew exponential power


        Examples
        --------
        None


        History
        -------
        Written,  JM, May 2016
    """

    if sig is not None: sca = sig

    sSEP = sample_sep01(nn, xi=xi, beta=beta)
    SEP  = loc + sca * sSEP

    return SEP


def sample_sep01(nn, xi=1., beta=0.):
    """
        Samples from the skew exponential power distribution with location zero and scale one.


        Definition
        ----------
        def sample_sep01(nn, xi=1., beta=0.):


        Input
        -----
        nn         number of samples


        Optional Input
        --------------
        xi         parameter which controls the skewness
        beta       parameter which controls the kurtosis


        Output
        ------
        Samples from the standardized skew exponential power distribution


        Examples
        --------
        None


        Literature
        --------
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.
        --> Steps (6) described on page 5


        History
        -------
        Written,  JM, May 2016
    """
    from pyjams.jams.distributions import sep_fs_mean, sep_fs_std

    SEP_fs = sample_sep01_fs(nn, xi=xi, beta=beta)

    # (6) Standardize SEP_fs
    mean_sep_fs = sep_fs_mean(xi=xi, beta=beta)
    std_sep_fs  = sep_fs_std(xi=xi, beta=beta)
    sSEP        = (SEP_fs - mean_sep_fs) / std_sep_fs   # standardized SEP (=Schoups and Vrugt's a_t)

    return sSEP


# --------------------------------------------------------------------


def sample_sep_fs(nn, loc=0., sca=1., xi=1., beta=0., sig=None):
    """
        Samples from the skew exponential power distribution with
        given location, scale, and parameters controlling skewness and kurtosis,
        using the approach of Fernandez and Steel (1998).

        The xi parameter needs to be positive.
             1 = symmetric
            >1 = right skewed
            <1 = left skewed

        The beta parameter needs to be from [-1,1]
             1 = fat tail
             0 = medium tails
            -1 = thin tail

        If xi = 1 and beta = 0, the SEP is the Gaussian distribution.


        Definition
        ----------
        def sample_sep_fs(nn, loc=0., sca=1., xi=1., beta=0., sig=None):


        Input
        -----
        nn         number of samples


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         parameter which controls the skewness
        beta       parameter which controls the kurtosis
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples from the (general) skew exponential power


        Examples
        --------
        None


        History
        -------
        Written,  JM, May 2016
    """

    if sig is not None: sca = sig

    sSEP = sample_sep01_fs(nn, xi=xi, beta=beta)
    SEP  = loc + sca * sSEP

    return SEP


def sample_sep01_fs(nn, xi=1., beta=0.):
    """
        Samples from the skew exponential power distribution with location zero and scale one and
        parameters controlling skewness and kurtosis,
        using the approach of Fernandez and Steel (1998).

        If xi = 1. then mean=0. and stddev=1.


        Definition
        ----------
        def sample01_sep_fs(nn, xi=1., beta=0.):


        Input
        -----
        nn       -> number of samples


        Optional Input
        --------------
        xi         parameter which controls skewness
        beta       parameter which controls the kurtosis


        Output
        ------
        Samples from the skew exponential power distribution


        Examples
        --------
        None


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.
        --> Steps (4)-(5) described on page 5


        History
        -------
        Written,  JM, May 2016
    """

    # (1) - (3)
    EP = sample_ep01(nn, beta=beta)

    # (4) Generate a random sign w_t (+1 or -1) with probabilities 1-xi/(xi+1/xi) and xi/(xi+1/xi)
    #     wt = np.where(np.random.rand(nn) > 1.-xi/(xi+1./xi), 1, -1)
    # (5) Compute SEP_t which is a sample from the skew exponential power distribution
    #     SEP(mu_xi, sigma_xi, xi, beta) with mu_xi and sigma_xi given by (A5) and (A6)
    #     SEPt = wt * np.abs(EPt) * xi**wt
    SEP_fs = skew_fernandez_steel(EP, xi)

    return SEP_fs



# --------------------------------------------------------------------



def sample_st(nn, nu, loc=0., sca=1., xi=1., sig=None):
    """
        Samples from the skew Student-t distribution with
        given location, scale, and parameter controlling skewness.

        The parameter controlling skewness xi needs to be positive.
             1 = symmetric
            >1 = right skewed
            <1 = left skewed


        Definition
        ----------
        def sample_st(nn, nu, loc=0., sca=1., xi=1., sig=None):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         parameter controlling skewness
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples from the skew Student-t distribution


        Examples
        --------
        None


        History
        -------
        Written,  JM, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)

    sst = sample_st01(nn, nu, xi=xi)
    st  = loc + sca * sst

    return st


def sample_st01(nn, nu, xi=1.):
    """
        Samples from the skew Student-t distribution with loc=0 and scale=1.


        Definition
        ----------
        def sample_st01(nn, nu, xi=1.):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Optional Input
        --------------
        xi         parameter controlling skewness


        Output
        ------
        Samples from the skew Student-t distribution iwth loc=0, sca=1


        Examples
        --------
        None


        History
        -------
        Written,  JM, May 2016
    """
    from pyjams.jams.distributions import st_fs_mean, st_fs_std

    st_fs = sample_st_fs(nn, nu, xi=xi)

    # (6) Standardize st_fs
    mean_st_fs = st_fs_mean(nu, xi=xi)
    std_st_fs  = st_fs_std(nu, xi=xi) * np.sqrt((nu-2.)/nu)
    sst        = (st_fs - mean_st_fs) / std_st_fs   # standardized skew Student-t

    return sst


# --------------------------------------------------------------------


def sample_st_fs(nn, nu, loc=0., sca=1., xi=1., sig=None):
    """
        Samples from the skew Student-t distribution with
        given location, scale, and parameter controlling skewness,
        using the approach of Fernandez and Steel (1998).

        The parameter controlling skewness xi needs to be positive.
             1 = symmetric
            >1 = right skewed
            <1 = left skewed


        Definition
        ----------
        def sample_st_fs(nn, nu, loc=0., sca=1., xi=1., sig=None):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         parameter controlling skewness
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples from the skew Student-t distribution


        Examples
        --------
        None


        History
        -------
        Written,  JM, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)

    sst = sample_st01_fs(nn, nu, xi=xi)
    st  = loc + sca * sst

    return st


def sample_st01_fs(nn, nu, xi=1.):
    """
        Samples from the skew Student-t distribution obtained
        by using the approach of Fernandez and Steel (1998).


        Definition
        ----------
        def sample_st01_fs(nn, nu, xi=1.):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Optional Input
        --------------
        xi       parameter controlling skewness


        Output
        ------
        Samples from the skew Student-t distribution after Fernandez and Steel (1998).


        Examples
        --------
        None


        Literature
        --------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  JM, May 2016
    """

    # (1) - (3)
    t = sample_t01(nn, nu)

    # (4) Generate a random sign w_t (+1 or -1) with probabilities 1-xi/(xi+1/xi) and xi/(xi+1/xi)
    #     wt = np.where(np.random.rand(nn) > 1.-xi/(xi+1./xi), 1, -1)
    # (5) Compute SEP_t which is a sample from the skew exponential power distribution
    #     SEP(mu_xi, sigma_xi, xi, beta) with mu_xi and sigma_xi given by (A5) and (A6)
    #     SEPt = wt * np.abs(EPt) * xi**wt
    st_fs = skew_fernandez_steel(t, xi)

    return st_fs


# --------------------------------------------------------------------


def sample_t(nn, nu, loc=0., sca=1., sig=None):
    """
        Samples from the Student-t distribution t(nu,loc,sca) are drawn.

        The mean is loc, standard deviation is sca*np.sqrt(nu/(nu-2)).

        The distribution is symmetric.


        Definition
        ----------
        def sample_t(nn, nu, loc=0., sca=1., sig=None):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        sig        standard deviation, overwrites sca


        Output
        ------
        Samples drawn from Student-t distribution


        Examples
        --------
        None


        History
        -------
        Written  JM, May 2016
    """
    # import scipy.stats as ss

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)

    t01 = sample_t01(nn, nu)
    t   = t01 * sca + loc
    # t = ss.t.rvs(nu, loc=loc, scale=sca, size=nn)

    return t


def sample_t01(nn, nu):
    """
        Samples from the Student-t distribution with loc=0 and scale=1 t(df=nu, loc=0, sca=1) are drawn.

        The distribution is symmetric.


        Definition
        ----------
        def sample_t01(nn, nu):


        Input
        -----
        nn       -> number of samples
        nu       -> degrees of freedom


        Output
        ------
        Samples drawn from Student-t distribution with loc=0, scale=1.


        Examples
        --------
        None


        History
        -------
        Written  JM, May 2016
    """
    import scipy.stats as ss

    t01 = ss.t.rvs(nu, loc=0.0, scale=1.0, size=nn)

    return t01


# ------------------------------------------------------------


def skew_fernandez_steel(sample_symmetric, xi):
    """
        Skews any given symmetric sample with a skewness factor xi
        following the approch of Fernandez & Steel.


        Definition
        ----------
        def skew_fernandez_steel(sample_symmetric, xi):


        Input
        -----
        sample_symmetric       -> sample drawn from a symmetric distribution
        xi                     -> skewness factor


        Output
        ------
        Samples from the skewed version of symmetric distribution used to generate sample


        Examples
        --------
        None


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  JM, May 2016
    """

    rr        = np.random.rand(np.size(sample_symmetric))
    absSymm   = np.abs(sample_symmetric)
    sample_fs = np.where(rr > 1.-xi/(xi+1./xi), absSymm*xi, -absSymm/xi)

    if np.iterable(sample_symmetric):
        return sample_fs
    else:
        return float(sample_fs)


# ------------------------------------------------------------


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # # ------------------------------------------------------------
    # # Do some tests for the sampling of distributions
    # # ------------------------------------------------------------

    # import scipy.special as sp
    # import scipy.stats as ss
    # import jams
    # import matplotlib.pylab as plt
    # plt.close("all")

    # nn    = 10000000   # number of random samples drawn
    # loc   = 0.0       # location parameter = mean
    # sca   = 2.0       # scale parameer = standard deviation
    # xi    = 1.5       # skew parameter (0,inf) 1=symmetric, >1=right skewed, <1=left skewed
    # beta  = -0.5       # ex. kurtosis parameter (-1,1] 1=fat tail, -1=thin tail
    # nu    = 6        # degrees of freedom for student-t

    # test_ep  = True
    # test_sep = True
    # test_t   = True
    # test_st_fs = True
    # test_st  = True

    # print('loc  = ', loc)
    # print('sca  = ', sca)
    # print('xi   = ', xi)
    # print('beta = ', beta)
    # print('nu   = ', nu)
    # print('')

    # if test_ep:
    #     # --------------
    #     # EP tests
    #     # --------------
    #     print('')
    #     print('-----------------------------------------------------')
    #     ep_samples = jams.distributions.sample_ep(nn, loc=loc, sca=sca, beta=beta)
    #     plt.figure()
    #     plt.hist(ep_samples,bins=100,normed=True)
    #     print('EP:  mean_samp      = ', np.mean(ep_samples))
    #     print('EP:  std_samp       = ', np.std(ep_samples))
    #     print('EP:  skew_samp      = ', ss.skew(ep_samples))
    #     print('EP:  kurt_samp      = ', ss.kurtosis(ep_samples, fisher=True))
    #     print('')

    #     # Compare with pdf of EP
    #     dx = 0.01
    #     ep_xx = np.arange(-10,10,dx)
    #     ep_yy = jams.distributions.ep(ep_xx, loc=loc, sca=sca, beta=beta)

    #     mean_num = np.sum(ep_xx*ep_yy*dx)
    #     var_num  = np.sum((ep_xx-mean_num)**2*ep_yy*dx)
    #     std_num  = np.sqrt(var_num)
    #     skew_num  = np.sum(((ep_xx-mean_num)/std_num)**3*ep_yy*dx)
    #     kurt_num  = np.sum(((ep_xx-mean_num)/std_num)**4*ep_yy*dx) - 3.0   # Fisher definition (Gaussian kurt=0.0)
    #     print('EP:  mean_num       = ', mean_num )
    #     print('EP:  std_num        = ', std_num  )
    #     print('EP:  skew_num       = ', skew_num )
    #     print('EP:  kurt_num       = ', kurt_num )
    #     print('')

    #     bb = 2./(beta+1.)
    #     print('EP:  calc kurt      = ', (sp.gamma(5./bb)*sp.gamma(1./bb))/((sp.gamma(3./bb))**2) - 3. )

    #     plt.plot(ep_xx, ep_yy, 'r-', lw=5, alpha=0.6)
    #     plt.show()

    # if test_sep:
    #     # --------------
    #     # SEP tests
    #     # --------------
    #     print('')
    #     print('-----------------------------------------------------')
    #     sep_samples = jams.distributions.sample_sep(nn, loc=loc, sca=sca, xi=xi, beta=beta)
    #     plt.figure()
    #     plt.hist(sep_samples,bins=100,normed=True)
    #     print('SEP: mean_samp      = ', np.mean(sep_samples))
    #     print('SEP: std_samp       = ', np.std(sep_samples))
    #     print('SEP: skew_samp      = ', ss.skew(sep_samples))
    #     print('SEP: kurt_samp      = ', ss.kurtosis(sep_samples, fisher=True))
    #     print('')

    #     # Compare with pdf of SEP
    #     dx = 0.01
    #     sep_xx = np.arange(-10,10,dx)
    #     sep_yy = jams.distributions.sep(sep_xx, loc=loc, sca=sca, xi=xi, beta=beta)

    #     mean_num = np.sum(sep_xx*sep_yy*dx)
    #     var_num  = np.sum((sep_xx-mean_num)**2*sep_yy*dx)
    #     std_num  = np.sqrt(var_num)
    #     skew_num  = np.sum(((sep_xx-mean_num)/std_num)**3*sep_yy*dx)
    #     kurt_num  = np.sum(((sep_xx-mean_num)/std_num)**4*sep_yy*dx) - 3.0   # Fisher definition (Gaussian kurt=0.0)
    #     print('SEP: mean_num       = ', mean_num )
    #     print('SEP: std_num        = ', np.sqrt(var_num) )
    #     print('SEP: skew_num       = ', skew_num )
    #     print('SEP: kurt_num       = ', kurt_num )
    #     print('')

    #     plt.plot(sep_xx, sep_yy, 'r-', lw=5, alpha=0.6)
    #     plt.show()


    # if test_t:
    #     # --------------
    #     # Student-t tests
    #     # --------------
    #     print('')
    #     print('-----------------------------------------------------')
    #     t_samples = jams.distributions.sample_t(nn, nu, loc=loc, sca=sca)
    #     plt.figure()
    #     plt.hist(t_samples,bins=100,normed=True)
    #     print('Student-t:  mean_samp      = ', np.mean(t_samples),                 ' ~ ',loc,                    ' : theoretical')
    #     print('Student-t:  std_samp       = ', np.std(t_samples),                  ' ~ ',sca*np.sqrt(nu/(nu-2.)),' : theoretical')
    #     print('Student-t:  skew_samp      = ', ss.skew(t_samples),                 ' ~ ',0.0,                    ' : theoretical')
    #     print('Student-t:  kurt_samp      = ', ss.kurtosis(t_samples, fisher=True))
    #     print('')

    #     # Compare with pdf of Student-t
    #     dx = 0.0001
    #     t_xx = np.arange(-50+loc,50+loc,dx)
    #     t_yy = jams.distributions.t(t_xx, nu, loc=loc, sca=sca)

    #     mean_num = np.sum(t_xx*t_yy*dx)
    #     var_num  = np.sum((t_xx-mean_num)**2*t_yy*dx)
    #     std_num  = np.sqrt(var_num)
    #     skew_num  = np.sum(((t_xx-mean_num)/std_num)**3*t_yy*dx)
    #     kurt_num  = np.sum(((t_xx-mean_num)/std_num)**4*t_yy*dx) - 3.0   # Fisher definition (Gaussian kurt=0.0)
    #     print('Student-t:  mean_num       = ', mean_num )
    #     print('Student-t:  std_num        = ', std_num  )
    #     print('Student-t:  skew_num       = ', skew_num )
    #     print('Student-t:  kurt_num       = ', kurt_num )
    #     print('')

    #     plt.plot(t_xx, t_yy, 'r-', lw=5, alpha=0.6)
    #     plt.show()

    # if (test_st_fs):
    #     st_fs_samp = jams.distributions.sample_st_fs(nn, nu, xi=xi)
    #     mean_st_fs  = jams.distributions.st_fs_mean(nu, xi=xi)
    #     std_st_fs   = jams.distributions.st_fs_std(nu, xi=xi)

    #     print('ssStudent-t:  mean_func      = ', mean_st_fs)
    #     print('ssStudent-t:  std_func       = ', std_st_fs)
    #     print('')
    #     print('ssStudent-t:  mean_samp      = ', np.mean(st_fs_samp))
    #     print('ssStudent-t:  std_samp       = ', np.std(st_fs_samp))

    # if (test_st):
    #     # --------------
    #     # skewed Student-t tests
    #     # --------------
    #     print('')
    #     print('-----------------------------------------------------')
    #     # st_samples = jams.distributions.sample_st(nn, nu, loc=loc, sig=sca*np.sqrt(nu/(nu-2.)), xi=xi)
    #     st_samples = jams.distributions.sample_st(nn, nu, loc=loc, sca=sca, xi=xi)
    #     plt.figure()
    #     plt.hist(st_samples,bins=100,normed=True)
    #     print('sStudent-t:  mean_samp      = ', np.mean(st_samples))
    #     print('sStudent-t:  std_samp       = ', np.std(st_samples))
    #     print('sStudent-t:  skew_samp      = ', ss.skew(st_samples))
    #     print('sStudent-t:  kurt_samp      = ', ss.kurtosis(st_samples, fisher=True))
    #     print('')

    #     # Compare with pdf of Student-t
    #     dx = 0.0001
    #     st_xx = np.arange(-30+loc,30+loc,dx)
    #     # st_yy = jams.distributions.st(st_xx, nu, loc=loc, sig=sca*np.sqrt(nu/(nu-2.)), xi=xi)
    #     st_yy = jams.distributions.st(st_xx, nu, loc=loc, sca=sca, xi=xi)

    #     mean_num = np.sum(st_xx*st_yy*dx)
    #     var_num  = np.sum((st_xx-mean_num)**2*st_yy*dx)
    #     std_num  = np.sqrt(var_num)
    #     skew_num  = np.sum(((st_xx-mean_num)/std_num)**3*st_yy*dx)
    #     kurt_num  = np.sum(((st_xx-mean_num)/std_num)**4*st_yy*dx) - 3.0   # Fisher definition (Gaussian kurt=0.0)
    #     print('sStudent-t:  mean_num       = ', mean_num )
    #     print('sStudent-t:  std_num        = ', std_num  )
    #     print('sStudent-t:  skew_num       = ', skew_num )
    #     print('sStudent-t:  kurt_num       = ', kurt_num )
    #     print('')

    #     plt.plot(st_xx, st_yy, 'r-', lw=5, alpha=0.6)
    #     plt.show()



#!/usr/bin/env python
import numpy as np
import pyjams


__all__ = ['get_manual_flags']


# --------------------------------------------------------------------

def get_manual_flags(flagfile, variable, dendro=None, julian=True):
    """
        Get start and end dates as well as flag values for a specific variable from a manual flag file.

        If start date is missing, 01.01.1900 00:00:00 will be used.
        If end date is missing, 01.01.2099 23:59:59 will be used.
        If no manual flag is found, empty lists wil be returned.

        The manual flag files is supposed to have the following structure
        var_id,start,end,flag,comment
        Lines starting with # will be ignored.


        Input
        -----
        flagfile   Filename of manual flag file
        variable   variable name in first column of flagfile


        Optional Input
        --------------
        julian     True: return julian dates (default)
                   False: return ascii strings in format DD.MM.YYYY hh:mm
        dendro     True: return also diameter at breast height and corresponding dendrometer reading
                   False: return only dates and flag


        Output
        ------
        list of start dates, list of end dates, list of flags (if dendro=False)
        list of start dates, list of end dates, list of flags, list of diameter at breast height, list of initial dendrometer readings (if dendro=False)


        Examples
        --------
        --> see __init__.py for full example of workflow

        sdate, edate, mflags = get_manual_flags('Manual_Flags_Soilnet_HH.csv"', 'Box02_Moist1')
        sdate, edate, mflags, bdh, d_ini = get_manual_flags('Manual_Flags_Soilnet_HH.csv"', 'Box02_Moist1', dendro=True)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2016 Matthias Cuntz, Benjamin Dechant - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Aug 2015
        Modified, AW, Aug 2015, optional readout of dendrometer-data:  DBH and Dini
        Modified, MC, Jan 2016, igrnore leading and trailing balnk in variable ids
        Modified, BD, Oct 2016, corrected and improved error message if flag missing; introduced error messages if sdate or edate are not in the correct format
    """
    # Default dates
    sdef = "01.01.1900 00:00:00"
    edef = "01.01.2099 23:59:59"

    # Read manual flag file
    sdat    = pyjams.sread(flagfile, comment="#", strarr=True, skip=0)
    sdat1 = sdat[0,:]      #header
    sdat = sdat[1:,:]      # get rid of header for the rest
    var_id  = np.array([ i.strip() for i in sdat[:,0] ])
    start   = sdat[:,1]
    end     = sdat[:,2]
    flag    = sdat[:,3]
    if dendro:
        dbh     = sdat[:,4]
        ii = np.where(dbh=='')[0]
        if ii.size > 0: dbh[ii] = '0'
        d_ini   = sdat[:,5]
        ii = np.where(d_ini=='')[0]
        if ii.size > 0: d_ini[ii] = '0'
    # comment = sdat[:,-1]

    # select variable
    ii = np.where(var_id == variable)[0]
    if ii.size > 0:
        sdate   = start[ii]
        edate   = end[ii]
        try:
            mflag   = flag[ii].astype(np.int)
        except:
            miss_ind = [ind for ind,fl in enumerate(flag[ii]) if fl == ''][0]
            comm_col = [m for m,com in enumerate(sdat1) if com[0:7] == 'comment']
            comm = sdat[ii,comm_col][miss_ind]
            print('get_manual_flags: forgotten flag in file '+flagfile+' - var/start/end: '+variable+'/'+str(sdate[miss_ind])+'/'+str(edate[miss_ind])+' - comment: '+comm)
            pyjams.jams.encrypt.sendfail(
                'get_manual_flags: forgotten flag in file '+flagfile+' - var/start/end: '+variable+'/'+str(sdate[miss_ind])+'/'+str(edate[miss_ind])+' - comment: '+comm,
                sender='benjamin.dechant@ufz.de')
        if dendro:
            l_dbh   = dbh[ii].astype(np.float)
            l_d_ini = d_ini[ii].astype(np.float)
        # Fill default dates
        jj = np.where(sdate == "")[0]
        if jj.size > 0: sdate[jj] = sdef
        jj = np.where(edate == "")[0]
        if jj.size > 0: edate[jj] = edef
        # Julian or ascii
        if julian:
            try:
                sdate = pyjams.jams.date2dec(ascii=sdate)
            except Exception as e:
                for sda_ind,sda in enumerate(sdate):
                    try:
                        sdate = pyjams.jams.date2dec(ascii=sda)
                    except:
                        pb_ind = sda_ind
                comm_col = [m for m,com in enumerate(sdat1) if com[0:7] == 'comment']
                comm = sdat[ii,comm_col][pb_ind]
                print('get_manual_flags: invalid start date in file '+flagfile+' - var/start: '+variable+'/'+str(e)[-19:-1]+' - comment: '+comm)
                pyjams.jams.encrypt.sendfail(
                    'get_manual_flags: invalid start date in file '+flagfile+' - var/start: '+variable+'/'+str(e)[-19:-1]+' - comment: '+comm,
                    sender='benjamin.dechant@ufz.de')
                raise ValueError

            try:
                edate = pyjams.jams.date2dec(ascii=edate)
            except Exception as e:
                for sda_ind,sda in enumerate(edate):
                    try:
                        sdate = pyjams.jams.date2dec(ascii=sda)
                    except:
                        pb_ind = sda_ind
                comm_col = [m for m,com in enumerate(sdat1) if com[0:7] == 'comment']
                comm = sdat[ii,comm_col][pb_ind]
                print('get_manual_flags: invalid end date in file '+flagfile+' - var/end: '+variable+'/'+str(e)[-19:-1]+' - comment: '+comm)
                pyjams.jams.encrypt.sendfail(
                    'get_manual_flags: invalid end date in file '+flagfile+' - var/end: '+variable+'/'+str(e)[-19:-1]+' - comment: '+comm,
                    sender='benjamin.dechant@ufz.de')
                raise ValueError                        # maybe not necessary

    else:
        # return empty lists if variable not in manual flag file
        sdate = list()
        edate = list()
        mflag = list()
        if dendro:
            l_dbh   = list()
            l_d_ini = list()

    if dendro:
        return [sdate, edate, mflag, l_dbh, l_d_ini]
    else:
        return [sdate, edate, mflag]



# --------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # base_folder = '/Volumes/Gruppen/CHS-Data'
    # data_folder = base_folder+'/HohesHolz'
    # location_short_name = 'HH'
    # man_file = data_folder+"/Soilnet/Manual_Flags_Soilnet_"+location_short_name+".csv"

    # v = 'Box02_Moist1'
    # print(v)
    # s, e, m = get_manual_flags(man_file, v)
    # for i in range(len(s)):
    #     print(s[i], e[i], m[i])
    # s, e, m = get_manual_flags(man_file, v, julian=False)
    # for i in range(len(s)):
    #     print(s[i], e[i], m[i])

    # v = 'Box27_Moist1'
    # print(v)
    # s, e, m = get_manual_flags(man_file, v, julian=False)
    # for i in range(len(s)):
    #     print(s[i], e[i], m[i])



#!/usr/bin/env python
"""
This is the unittest for the Logistic Functions module.

python -m unittest -v tests/test_logistic_function.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_logistic_function.py

"""
import unittest


class TestLogisticFunction(unittest.TestCase):
    """
    Tests for functions/logistic_function.py

    """

    def test_logistic_function(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import logistic, logistic_offset
        from pyjams.functions import logistic2_offset
        from pyjams.functions import dlogistic, dlogistic_offset
        from pyjams.functions import dlogistic2_offset
        from pyjams.functions import d2logistic, d2logistic_offset
        from pyjams.functions import d2logistic2_offset
        from pyjams.functions import logistic_p, logistic_offset_p
        from pyjams.functions import logistic2_offset_p
        from pyjams.functions import dlogistic_p, dlogistic_offset_p
        from pyjams.functions import dlogistic2_offset_p
        from pyjams.functions import d2logistic_p, d2logistic_offset_p
        from pyjams.functions import d2logistic2_offset_p

        assert logistic(1., 1., 0., 2.) == 0.5
        assert logistic(1., 1., 2., 1.) == 0.5
        assert logistic(2., 1., 1., 1.) == 1. / (1. + np.exp(-1.))
        assert logistic_offset(1., 1., 0., 2., 1.) == 1.5
        assert logistic_offset(1., 1., 2., 1., 1.) == 1.5
        assert (logistic_offset(2., 1., 1., 1., 1.) ==
                1. / (1. + np.exp(-1.)) + 1.)
        assert logistic2_offset(1., 1., 2., 1., 2., 2., 1., 1.) == 0.5
        assert dlogistic(1., 1., 2., 1.) == 0.5
        assert dlogistic_offset(1., 1., 2., 1., 1.) == 0.5
        assert dlogistic2_offset(1., 1., 2., 1., 2., 2., 1., 1.) == -0.5
        assert np.around(d2logistic(1., 1., 2., 2.), 4) == 0.3199
        assert np.around(d2logistic_offset(1., 1., 2., 2., 1.), 4) == 0.3199
        assert (np.around(d2logistic2_offset(1., 1., 2., 2., 2., 2., 2., 1.),
                          4) == -0.3199)

        assert logistic_p(1., [1., 0., 2.]) == 0.5
        assert logistic_p(1., [1., 2., 1.]) == 0.5
        assert logistic_p(2., [1., 1., 1.]) == 1. / (1. + np.exp(-1.))
        assert logistic_offset_p(1., [1., 0., 2., 1.]) == 1.5
        assert logistic_offset_p(1., [1., 2., 1., 1.]) == 1.5
        assert (logistic_offset_p(2., [1., 1., 1., 1.]) ==
                1. / (1. + np.exp(-1.)) + 1.)
        assert logistic2_offset_p(1., [1., 2., 1., 2., 2., 1., 1.]) == 0.5
        assert dlogistic_p(1., [1., 2., 1.]) == 0.5
        assert dlogistic_offset_p(1., [1., 2., 1., 1.]) == 0.5
        assert dlogistic2_offset_p(1., [1., 2., 1., 2., 2., 1., 1.]) == -0.5
        assert np.around(d2logistic_p(1., [1., 2., 2.]), 4) == 0.3199
        assert (np.around(d2logistic_offset_p(1., [1., 2., 2., 1.]), 4)
                == 0.3199)
        assert (np.around(
            d2logistic2_offset_p(1., [1., 2., 2., 2., 2., 2., 1.]),
            4) == -0.3199)

        self.assertEqual(list(logistic(np.array([1., 1.]), 1., 0., 2.)),
                         [0.5, 0.5])

        ll = logistic(pd.Series([1., 1.]), 1., 0., 2.)
        self.assertEqual(list(ll), [0.5, 0.5])
        ll = logistic(pd.DataFrame([[1., 1.], [1., 1.]]), 1., 0., 2.)
        self.assertEqual(list(ll.to_numpy().flatten()), [0.5, 0.5, 0.5, 0.5])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from pyjams.jams import eddybox as eb
from pyjams.jams.timestepcheck import timestepcheck
'''
    Example file for processing Eddy Covariance data with eddybox and EddySoft.


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written  AP, Sep 2014
'''

'''
0. Folder structure for eddy data processing:
e.g. |-2013                             :containing everything
       |-config                 :containg log and config files
       |-flux                   :for final fluxes
         |-itc                  :for itc plots
         |-spike                        :for spike plots
         |-ustar                        :for ustar plots and thresholds
       |-lags                   :for lags from eddycorr and plots
       |-meteo                  :containing meteo files
       |-online                 :containing online calculated files (unused)
       |-profile         :containing profile data
       |-pfit                   :for planar fit plots and files
       |-raw                            :for uncorrected raw fluxes
       |-slt                            :containing slt files
         |-deleted              :for slts to small
       |-spec                   :for inductance files and plots
       eddysuite.py             :this file
'''

'''
1. clean slt files
'''
#eb.sltclean('slt')

'''
2. calculate raw fluxes
Use EddyFlux to calculate raw, uncorrected fluxes
- no meteo files, no time lag, no filter, no inductances, no coord rotation
- no extra output, no spike detection, p=1000, set sonic direction correctly
- save results in raw/raw.csv
- save settings in raw/raw.cal
'''

'''
3. sync meteorological data with available *.slt data, requires the meteo file to be timestep checked
Use only full year meteo data with no missing time stamps
'''
#eb.meteo4slt('slt', 'meteo/meteo.csv', [1,2,6], 'meteo/meteo4slt.csv')
#eb.fluxplot('meteo/meteo.csv', 'meteo/meteo.pdf', units=False, plot=True)

'''
4. calculate time lags
Use EddyCorr to calculate time lags between w+c and w+h
- transform to meteorological data, every 1. value, scan rate 20 Hz
- lag: e.g. 300, the bigger, the longer it takes
- substract mean, extreme values
'''

'''
5. modell missing lags and attach lags to meteo file
You will be asked to give maximum, minimum and median lags. For moisture depdendend
modelling of the H2O lag, user is asked to give limits to used moisture values and
lags for fitting as well as initial guess for the model
'''
#eb.eddycorr('lags', 'slt', '35_corr.csv', '36_corr.csv', 'meteo/meteo4slt.csv', 'lags.csv', plot=True)

'''
6. Calculate inductances for carbon and water fluxes
Calculate inductances for multiple times within the year, e.g. ones per month
Use EddySpec in between:
- Co- and Quadrature spectra between w+t, w+c and w+h with the files shown
- transform to meteorological data, normalize to 1, NO inductances
- scan rate 20, dynamic smooth, multiply by freq., log.equid. freq., Welch
- for w+t lag=0, for w+c lag=c lag determined before (pos.), for w+h lag=h lag determined before (pos.)
Use SpecMean to average each spectrum for w+t, w+c and w+h
'''
#eb.eddyspec('spec', 'lags/35_corr.csv', 'lags/36_corr.csv', 'raw/raw.csv', 'slt', plot=True)

'''
7. Do the planar fit rotation for entire year or vegetations seasons
Use EddyPFit in between with:
- file format csv, minimum 100/sector, start at 0, width 30 deg, upper limit 1 m/s
- lower limit .1 m/s
'''
#eb.planarfit('pfit', 'raw/raw.csv', '2013.csv', plot=True)

'''
8. Use EddyFlux to calculate fluxes with all corrections
- Enable spike detection in EddyConf
- use inductances from spec folder, use PF, sect. with pfitmatrix file from pfit folder
- extra output, save the settings in flux
Do the timestepcheck and correct time stamps not in regular interval. Plot afterwards.
'''
#timestepcheck('flux', 'flux.csv', 'flux/flux_checked.csv', '01.01.2013 00:30', '01.01.2014 00:00')
#eb.fluxplot('flux/flux_checked.csv', 'flux/flux_checked.pdf', units=True, plot=True)

'''
9. Calculation of quality flags
- ustar flagging only possible with one full year data set
'''
swdr = 21       # column number of short wave downward radiation in meteo.csv
T    = 1        # column number of air temperature in meteo.csv
lat  = 50.450127 # latitude of tower position in decimal degrees
#eb.fluxflag('flux/flux_checked.csv', 'meteo/meteo.csv', 'flux', swdr, T, lat, spike_f=True, itc_f=True, ustar_f=True, plot=True)

'''
10. Calculation of storage fluxes
'''
#eb.profile2storage('flux/fluxflags.csv', 'flux/flux_checked.csv', 'profile/profile.csv', 'flux', heights=[0.1,0.3,1.0,2.0,5.0,9.0,15.0,23.8,30.0],
#                   CO2=[0,1,2,3,4,5,6,7,8], H2O=[9,10,11,12,13,14,15,16,17], delimiter=[',', ',', ';'], plot=True)
#eb.profile2storage('flux/flux+stor.csv', 'flux/flux_checked.csv', 'meteo/meteo.csv', 'flux', heights=[2.0,30.0], T=[2,1], plot=True)

'''
11. Gap filling
'''
rg = 19
tair = 1
rh = 5
#eb.fluxfill('flux/flux+stor.csv', 'meteo/meteo.csv', 'flux', rg, tair, rh, undef=-9999, plot=True)

'''
12. Partitioning
'''
swdr = 21
tair = 1
rh = 5
#eb.fluxpart('flux/fluxfilled.csv', 'meteo/meteo.csv', 'flux', swdr, tair, rh, method='reichstein', nogppnight=True, undef=-9999, plot=True)

'''
13. Energy balance closure
'''
swdr = 21
Rn = 32
G = 33
#eb.energyclosure('flux/fluxpart.csv', 'meteo/meteo.csv', 'flux', Rn, G, swdr, method='year', undef=-9999, plot=True)

'''
14. You're done! :-)
'''






#!/usr/bin/env python
import numpy as np
import netCDF4 as nc


def gapfill(date, data, rg, tair, vpd,
            data_flag=None, rg_flag=None, tair_flag=None, vpd_flag=None,
            rg_dev=50., tair_dev=2.5, vpd_dev=5.,
            longgap=60, fullday=False, undef=-9999, ddof=1,
            err=False, shape=False, verbose=0):
    """
        Fills gaps of flux data from Eddy covariance measurements according to
        Reichstein et al. (2005).
        If there is a gap in the data, look for similar meteorological conditions
        (defined as maximum possible deviations) in a certain time window and fill
        with the average of these 'similar' values.

        The routine can also do the same search for similar meteorological conditions
        for every data point and calculate its standard deviation as a measure of uncertainty.


        Definition
        ----------
        def gapfill(date, data, rg, tair, vpd,
                    data_flag=None, rg_flag=None, tair_flag=None, vpd_flag=None,
                    rg_dev=50., tair_dev=2.5, vpd_dev=5.,
                    longgap=60, fullday=False, undef=np.nan, ddof=1,
                    err=False, shape=False):


        Input
        -----
        date        1D-array of julian days
        data        ND-array of fluxes to fill
        rg          1D-array of global radiation [W m-2]
        tair        1D-array of air Temperature [deg C]
        vpd         1D-array of vapour pressure deficit [hPa]


        Optional Input
        -------------
        data_flag     1D-array of flags of fluxes: False=good quality; True=bad data (default: False)
        rg_flag       1D-array of flags of global radiation: False=good quality; True=bad data (default: False)
        tair_flag     1D-array of flags of air temperature: False=good quality; True=bad data (default: False)
        vpd_flag      1D-array of flags of vapour pressure deficit: False=good quality; True=bad data (default: False)


        Parameters
        ----------
        rg_dev               threshold for maximum deviation of global radiation (default: 50)
        tair_dev             threshold for maximum deviation of air Temperature (default: 2.5)
        vpd_dev              threshold for maximum deviation of vpd (default: 5)
        longgap              avoid extraploation into a gap longer than longgap days (default: 60)
        fullday              if True, move beginning of large gap to start of next day
                                      and move end of large gap to end of last day (default: False)
        undef                undefined values in data  (default: -9999, np.nan not allowed)
        ddof                 degrees of freedom to use in calculation of standard deviation
                             for error estimate (default: 1)
        err                  if True, fill every data point, i.e. used for error generation (default: False)
        shape                if False then outputs are 1D arrays;
                             if True, output have the same shape as input data;
                             if a tuple is given, then this tuple is used to reshape. (default: False)


        Ouput
        -----
        if err=False:
            filled_data, quality_class
        else:
            err_data


        Restrictions
        ------------
        If err=True, there is no error estimate if there are no meteorological
          conditions in the vicinity of the data point.
        Works not with undef=np.nan


        Literature
        ----------
        Reichstein et al. (2005)
            On the separation of net ecosystem exchange into assimilation and ecosystem
            respiration: review and improved algorithm.
            Global Change Biology 11, 1424-1439


        Examples
        --------
        >>> import numpy as np
        >>> from jams.fread import fread
        >>> from jams.date2dec import date2dec
        >>> from jams.autostring import astr
        >>> ifile = 'test_gapfill.csv' # Tharandt 1998 = Online tool example file
        >>> undef = -9999.
        >>> dat   = fread(ifile, skip=2, transpose=True)
        >>> ndat  = dat.shape[1]
        >>> head  = fread(ifile, skip=2, header=True)
        >>> head1 = head[0]
        >>> ihead = dict(list(zip(head1, list(range(len(head1))))))
        >>> for ii in range(len(head1)): exec(head1[ii].lower() + ' = ' + 'dat[ihead["'+head1[ii]+'"],:]')
        >>> year  = np.ones(day.shape, dtype=day.dtype) * 1998.
        >>> hh    = hour.astype(int)
        >>> mn    = np.round((hour-hh)*60.)
        >>> y0    = date2dec(yr=year[0], mo=1, dy=1, hr=hh, mi=mn)
        >>> jdate = y0 + day

        # 1D fill
        >>> nee_f, nee_qc = gapfill(jdate, nee, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True)
        >>> print(astr(nee_qc[11006:11012],0,pp=True))
        ['1' '1' '1' '2' '2' '2']
        >>> print(astr(nee_f[11006:11012],2,pp=True))
        ['-18.68' '-15.63' '-19.61' '-15.54' '-12.40' '-15.33']

        # 1D err
        >>> nee_std = gapfill(jdate, nee, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True, err=True)
        >>> print(astr(nee_std[11006:11012],3,pp=True))
        ['    5.372' '   13.118' '    6.477' '-9999.000' '-9999.000' '-9999.000']

        >>> nee_err     = np.ones(nee_std.shape, dtype=int)*(-1)
        >>> kk          = np.where((nee_std!=undef) & (nee_f!=0.))[0]
        >>> nee_err[kk] = np.abs(nee_std[kk]/nee_f[kk]*100.).astype(int)
        >>> print(astr(nee_err[11006:11012],pp=True))
        [' 28' ' 83' ' 33' ' -1' ' -1' ' -1']

        # 2D fill
        >>> nee2 = np.dstack([nee,nee,nee])
        >>> print(nee2.shape)
        (1, 17520, 3)
        >>> nee_f2, nee_qc2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True)
        >>> print(nee_f2.shape, nee_qc2.shape)
        (1, 17520, 3) (17520,)
        >>> nee_f, nee_qc   = gapfill(jdate, nee,  rg, tair, vpd, data_flag=(qcnee>1), undef=undef)
        >>> nee_f2, nee_qc2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef)
        >>> print(nee_f2.shape, nee_qc2.shape)
        (17520, 3) (17520,)
        >>> print(np.all(nee_f==nee_f2[:,0]), np.all(nee_f==nee_f2[:,1]), np.all(nee_f==nee_f2[:,2]))
        True True True

        # 2D err
        >>> nee_std  = gapfill(jdate, nee,  rg, tair, vpd, data_flag=(qcnee>1), undef=undef, err=True)
        >>> nee_std2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, err=True)
        >>> print(np.all(nee_std==nee_std2[:,0]), np.all(nee_std==nee_std2[:,1]), np.all(nee_std==nee_std2[:,2]))
        True True True


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2012 - modified gap_filling.py
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Apr 2014 - assert
                               - data ND-array
                               - longestmarginalgap was only working at beginning and end of time series
                                 renamed to longgap
                               - fullday
    """

    # -------------------------------------------------------------
    # Checks

    # check input types
    tma = np.ma.core.MaskedArray
    assert type(date) != tma, 'dates cannot be masked array.'
    if type(data) == tma: data = data.filled(undef)
    if type(rg)   == tma: rg   = rg.filled(undef)
    if type(tair) == tma: tair = tair.filled(undef)
    if type(vpd)  == tma: vpd  = vpd.filled(undef)

    # remember shape if any
    inshape = data.shape
    date = date.squeeze()
    data = data.squeeze()
    rg   = rg.squeeze()
    tair = tair.squeeze()
    vpd  = vpd.squeeze()
    assert date.ndim == 1, 'squeezed dates must be 1D array.'
    assert rg.ndim   == 1, 'squeezed rg must be 1D array.'
    assert tair.ndim == 1, 'squeezed tair must be 1D array.'
    assert vpd.ndim  == 1, 'squeezed vpd must be 1D array.'
    assert data.ndim <= 2, 'squeezed data must be 1D or 2D array.'

    # make 2D array
    if data.ndim == 1:
        is2d = False
        data = data[:,np.newaxis]
    else:
        is2d = True

    # check flags
    ndata  = data.shape[0]
    ndata2 = data.shape[1]
    if (data_flag is not None):
        data_flg = data_flag.squeeze()
    else:
        data_flg = np.where(data[:,0] == undef, True, False)
    if (rg_flag is not None):
        rg_flg = rg_flag.squeeze()
    else:
        rg_flg = np.where(rg == undef, True, False)
    if (tair_flag is not None):
        tair_flg = tair_flag.squeeze()
    else:
        tair_flg = np.where(tair == undef, True, False)
    if (vpd_flag is not None):
        vpd_flg = vpd_flag.squeeze()
    else:
        vpd_flg = np.where(vpd == undef, True, False)

    if ((date.size != ndata) | (rg.size != ndata) | (tair.size != ndata) |
        (vpd.size != ndata) | (data_flg.size != ndata) | (rg_flg.size != ndata) |
        (tair_flg.size != ndata) | (vpd_flg.size != ndata)):
        raise ValueError('inputs must have the same size.')

    # -------------------------------------------------------------
    # Parameters

    # number of data points per week; basic factor of the time
    # window
    ddate    = np.diff(date)
    if np.any((ddate-ddate[0]) > 1e-4): # 1e-4 are ca 2.secs
        raise ValueError('dates must be equally spaced.')
    week    = int(np.around(7./ddate[0]))
    nperday = week // 7
    # hour = (np.array(np.floor((date-np.trunc(date))*24.), dtype=int) + 12) % 24
    # hour, mi = dec2date(date, hr=True, mi=True)
    tobj = nc.num2date(date-1721424., 'days since 0001-01-01 12:00:00', calendar='gregorian')
    hour = np.array([ i.hour for i in tobj ], dtype=int)
    mi   = np.array([ i.minute for i in tobj ], dtype=int)
    hour = hour + mi/60.
    day  = np.floor(date-0.5).astype(int)

    if err:
        # error estimate
        data_std  = np.ones(data.shape)*undef
    else:
        # array for filled values
        data_fill = data.copy()
        ii = np.where(data_flg)[0]
        if ii.size > 0: data_fill[ii,:] = undef
        # gap quality classes
        quality = np.zeros(ndata)

    if verbose > 0:
        if err:
            print('  Start error estimate')
        else:
            print('  Start filling')

    #------------------------------------------------------------------
    # Large margins

    # Check for large margins at beginning
    largegap    = np.zeros(ndata, dtype=bool)
    firstvalid  = np.amin(np.where(~data_flg)[0])
    lastvalid   = np.amax(np.where(~data_flg)[0])
    nn          = int(nperday*longgap)
    if firstvalid > nn:
        if verbose > 1: print('    Large margin at beginning: ', firstvalid)
        largegap[0:(firstvalid-nn)] = True
    if lastvalid < (ndata-nn):
        if verbose > 1: print('    Large margin at end: ', lastvalid-nn)
        largegap[(lastvalid+nn):] = True

    # search largegap - code from maskgroup.py
    index  = []
    length = []
    count  = 0
    for i in range(ndata):
        if i==0:
            if data_flg[i]:
                index += [i]
                count  = 1
        if i>0:
            if data_flg[i] and not data_flg[i-1]:
                index += [i]
                count  = 1
            elif data_flg[i]:
                count += 1
            elif not data_flg[i] and data_flg[i-1]:
                length += [count]
                count = 0
            else:
                pass
    if count>0:
        length += [count]
    # set largegap
    for i in range(len(index)):
        if length[i] > nn:
            if verbose > 1: print('    Large gap: ', index[i], ':', index[i]+length[i])
            largegap[index[i]:index[i]+length[i]] = True

    # set or unset rest of days in large gaps
    if fullday:
        for i in range(ndata-1):
            if largegap[i] and not largegap[i+1]:   # end of large margin
                largegap[np.where(day == day[i])[0]] = False
            elif not largegap[i] and largegap[i+1]: # beginning of large margin
                largegap[np.where(day == day[i])[0]] = False
            else:
                continue

    #------------------------------------------------------------------
    # Gap filling

    # flag for all meteorological conditions
    meteo_flg  = (~tair_flg) & (~vpd_flg) & (~rg_flg)
    # flag for all meteorological conditions and data
    total_flag = meteo_flg & (~data_flg)

    # Fill loop over all data points
    for j in range(ndata):
        if not err:
            # no reason to go further, no gap -> continue
            if (~(data_flg[j])) | largegap[j]: continue
        # 3 Methods
        #   1. tair, vpd and global radiation;
        #   2. just global radiation;
        #   3. none meteorolgical conditions: take the mean of +- hour

        # for better overview: dynamic calculation of radiation threshold
        # minimum 20; maximum 50 [Wm-2] according to private correspondence
        # with Markus Reichstein
        rg_devmax = np.maximum(20.,np.minimum(rg[j],rg_dev))

        # Method 1: all met conditions
        if meteo_flg[j]:
            # search for values around the met-conditions in a window of time
            # (one week in the first iteration and odd weeks in the next)
            j1  = j - np.arange(1,week+1) + 1
            j2  = j + np.arange(1,week)
            jj  = np.append(j1,j2)
            win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
            # get boolean array where meteo-conditions are in a given width
            conditions = ( (np.abs(rg[win]  -rg[j])   < rg_devmax) &
                           (np.abs(tair[win]-tair[j]) < tair_dev) &
                           (np.abs(vpd[win] -vpd[j])  < vpd_dev) &
                           total_flag[win] )
            num4avg = np.sum(conditions)
            # we need at least two samples with similar conditions
            if num4avg >= 2:
                nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                dat = np.ma.array(data[win,:], mask=nconditions)
                if verbose > 2:
                    print('    m1.1: ', j, win.size, dat.mean(), dat.std(ddof=ddof))
                    # print(num4avg)
                    # print(conditions)
                    # print(np.sum(meteo_flg[win]))
                    # print(np.sum(~tair_flg[win]))
                    # print(tair[win][~tair_flg[win]])
                    # print(np.sum(~vpd_flg[win]))
                    # print(vpd[win][~vpd_flg[win]])
                    # print(np.sum(~rg_flg[win]))
                    # print(rg[win][~rg_flg[win]])
                    # print(np.sum(~data_flg[win]))
                    # print(np.sum(total_flag[win]))
                    # print(dat.compressed())
                if err:
                    data_std[j,:]  = np.ma.std(dat, axis=0, ddof=ddof)
                else:
                    data_fill[j,:] = np.ma.mean(dat, axis=0)
                    # assign also quality category of gap filling
                    quality[j]     = 1
                continue
            else: # --> extend time window to two weeks
                j1  = j - np.arange(1,2*week+1) + 1
                j2  = j + np.arange(1,2*week)
                jj  = np.append(j1,j2)
                win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
                conditions = ( (np.abs(rg[win]  -rg[j])   < rg_devmax) &
                               (np.abs(tair[win]-tair[j]) < tair_dev) &
                               (np.abs(vpd[win] -vpd[j])  < vpd_dev) &
                               total_flag[win] )
                num4avg = np.sum(conditions)
                if num4avg >= 2:
                    nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                    dat = np.ma.array(data[win,:], mask=nconditions)
                    if verbose > 2:
                        print('    m1.2: ', j, win.size, dat.mean(), dat.std(ddof=ddof))
                    if err:
                        data_std[j,:]  = np.ma.std(dat, axis=0, ddof=ddof)
                    else:
                        data_fill[j,:] = np.ma.mean(dat, axis=0)
                        quality[j]     = 1
                    continue

        # if you come here, no error estimate
        if err: continue

        # If nothing is found under similar meteo within two weeks,
        # look for global radiation within one week ->

        # Method 2: just global radiation available
        if (~(rg_flg[j])):
            j1  = j - np.arange(1,week+1) + 1
            j2  = j + np.arange(1,week)
            jj  = np.append(j1,j2)
            win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
            # get boolean array where meteo-conditions are in a given width
            conditions = ( (np.abs(rg[win]-rg[j]) < rg_devmax) &
                           total_flag[win] )
            num4avg = np.sum(conditions)
            # we need at least two samples with similar conditions
            if num4avg >= 2:
                nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                dat = np.ma.array(data[win,:], mask=nconditions)
                if verbose > 2:
                    print('    m2: ', j, win.size, dat.mean(), dat.std(ddof=ddof))
                if err:
                    data_std[j,:]  = np.ma.std(dat, axis=0, ddof=ddof)
                else:
                    data_fill[j,:] = np.ma.mean(dat, axis=0)
                    quality[j]     = 1
                continue

        # If still nothing is found under similar rg within one week,
        # take the same hour within 1-7 days

        # Method 3: same hour
        enough = False
        for i in range(2):
            t_win = (nperday * (2*i+1))//2
            j1  = j - np.arange(1,t_win+1) + 1
            j2  = j + np.arange(1,t_win)
            jj  = np.append(j1,j2)
            win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
            conditions = (np.abs(hour[win]-hour[j]) < 1.1) & (~(data_flg[win]))
            num4avg = np.sum(conditions)
            if num4avg >= 2:
                nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                dat = np.ma.array(data[win,:], mask=nconditions)
                if verbose > 2:
                    print('    m3.{:d}: '.format(i), j, win.size, dat.mean(), dat.std(ddof=ddof))
                if err:
                    data_std[j,:]  = np.ma.std(dat, axis=0, ddof=ddof)
                else:
                    data_fill[j,:] = np.ma.mean(dat, axis=0)
                    if i == 0:
                        quality[j] = 1
                    else:
                        quality[j] = 2
                break

        # sanity check
        if err:
            if data_std[j,0]  != undef: continue
        else:
            if data_fill[j,0] != undef: continue

        # If still nothing is found, start a new cycle with increased window size
        # Method 4: same as 1 but for 3-12 weeks
        if meteo_flg[j]:
            for multi in range(3,12):
                j1  = j - np.arange(1,multi*week+1) + 1
                j2  = j + np.arange(1,multi*week)
                jj  = np.append(j1,j2)
                win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
                conditions = ( (np.abs(rg[win]  -rg[j])   < rg_devmax) &
                               (np.abs(tair[win]-tair[j]) < tair_dev) &
                               (np.abs(vpd[win] -vpd[j])  < vpd_dev) &
                               total_flag[win] )
                num4avg = np.sum(conditions)
                # we need at least two samples with similar conditions
                if num4avg >= 2:
                    nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                    dat = np.ma.array(data[win,:], mask=nconditions)
                    if verbose > 2:
                        print('    m4.{:d}: '.format(multi), j, win.size, dat.mean(), dat.std(ddof=ddof))
                    if err:
                        data_std[j,:]  = np.ma.std(dat, axis=0, ddof=ddof)
                    else:
                        data_fill[j,:] = np.ma.mean(dat, axis=0)
                        # assign also quality category of gap filling
                        if multi <= 2:
                            quality[j] = 1
                        elif multi > 4:
                            quality[j] = 3
                        else:
                            quality[j] = 2
                    break
            # Check because continue does not support to jump out of two loops
            if err:
                if data_std[j,0]  != undef: continue
            else:
                if data_fill[j,0] != undef: continue

        # Method 5: same as 2 but for 2-12 weeks
        if (~(rg_flg[j])):
            for multi in range(2,12):
                j1  = j - np.arange(1,multi*week+1) + 1
                j2  = j + np.arange(1,multi*week)
                jj  = np.append(j1,j2)
                win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
                # get boolean array where meteo-conditions are in a given width
                conditions = ( (np.abs(rg[win]  -rg[j]) < rg_devmax) &
                               total_flag[win] )
                num4avg = np.sum(conditions)
                # we need at least two samples with similar conditions
                if num4avg >= 2:
                    nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                    dat = np.ma.array(data[win,:], mask=nconditions)
                    if verbose > 2:
                        print('    m5.{:d}: '.format(multi), j, win.size, dat.mean(), dat.std(ddof=ddof))
                    if err:
                        data_std[j,:] = np.ma.std(dat, axis=0, ddof=ddof)
                    else:
                        data_fill[j,:] = np.ma.mean(dat, axis=0)
                        if multi ==0:
                            quality[j] = 1
                        elif multi <= 2:
                            quality[j] = 2
                        else:
                            quality[j] = 3
                    break
            if err:
                if data_std[j,0]  != undef: continue
            else:
                if data_fill[j,0] != undef: continue

        # Method 6: same as 3 but for 3-120 days
        for i in range(3,120):
            t_win = nperday * (2*i+1)/2
            j1  = j - np.arange(1,t_win+1) + 1
            j2  = j + np.arange(1,t_win)
            jj  = np.append(j1,j2)
            win = np.unique(np.sort(np.clip(jj,0,ndata-1)))
            conditions = (np.abs(hour[win]-hour[j]) < 1.1) & (~(data_flg[win]))
            num4avg = np.sum(conditions)
            if num4avg >= 2:
                nconditions = np.repeat(~conditions,ndata2).reshape(conditions.size,ndata2)
                dat = np.ma.array(data[win,:], mask=nconditions)
                if verbose > 2:
                    print('    m6.{:d}: '.format(i), j, win.size, dat.mean(), dat.std(ddof=ddof))
                if err:
                    data_std[j,:] = np.ma.std(dat, axis=0, ddof=ddof)
                else:
                    data_fill[j,:] = np.ma.mean(dat, axis=0)
                    quality[j] = 3
                break

    if shape != False:
        if shape != True:
            ishape = shape
        else:
            ishape = inshape
        if err:
            return np.reshape(data_std,ishape)
        else:
            if np.prod(ishape) == quality.size: # 1D input with dimensions of 1
                return np.reshape(data_fill,ishape), np.reshape(quality,ishape)
            else:                               # 2D data but only 1D fill quality flags
                return np.reshape(data_fill,ishape), quality
    else:
        if err:
            if not is2d: data_std = data_std.squeeze()
            return data_std
        else:
            if not is2d: data_fill = data_fill.squeeze()
            return data_fill, quality


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np
    # from jams import fread
    # from jams import date2dec
    # from jams import astr
    # ifile = 'test_gapfill.csv' # Tharandt 1998 = Online tool example file
    # undef = -9999.
    # dat   = fread(ifile, skip=2, transpose=True)
    # # dat = dat[:,:1000]
    # ndat  = dat.shape[1]
    # head  = fread(ifile, skip=2, header=True)
    # head1 = head[0]
    # ihead = dict(list(zip(head1, list(range(len(head1))))))
    # for ii in range(len(head1)): exec(head1[ii].lower() + ' = ' + 'dat[ihead["'+head1[ii]+'"],:]')
    # year  = np.ones(day.shape, dtype=day.dtype) * 1998.
    # hh    = hour.astype(int)
    # mn    = np.round((hour-hh)*60.)
    # y0    = date2dec(yr=year[0], mo=1, dy=1, hr=hh, mi=mn)
    # jdate = y0 + day

    # # 1D fill
    # nee_f, nee_qc = gapfill(jdate, nee, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True, verbose=0)
    # print(astr(nee_qc[11006:11012],0,pp=True))
    # # ['1' '1' '1' '2' '2' '2']
    # print(astr(nee_f[11006:11012],3,pp=True))
    # # ['-18.678' '-15.633' '-19.610' '-15.536' '-12.402' '-15.329']

    # # 1D err
    # nee_std = gapfill(jdate, nee, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True, verbose=0, err=True)
    # print(astr(nee_std[11006:11012],3,pp=True))
    # # ['    5.372' '   13.118' '    6.477' '-9999.000' '-9999.000' '-9999.000']

    # nee_err     = np.ones(nee_std.shape, dtype=int)*(-1)
    # kk          = np.where((nee_std!=undef) & (nee_f!=0.))[0]
    # nee_err[kk] = np.abs(nee_std[kk]/nee_f[kk]*100.).astype(int)
    # print(astr(nee_err[11006:11012],pp=True))
    # # [' 28' ' 83' ' 33' ' -1' ' -1' ' -1']

    # # 2D fill
    # nee2 = np.dstack([nee,nee,nee])
    # print(nee2.shape)
    # # (1, 17520, 3)
    # nee_f2, nee_qc2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, shape=True)
    # print(nee_f2.shape, nee_qc2.shape)
    # # (1, 17520, 3) (17520,)
    # nee_f, nee_qc   = gapfill(jdate, nee,  rg, tair, vpd, data_flag=(qcnee>1), undef=undef)
    # nee_f2, nee_qc2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef)
    # print(nee_f2.shape, nee_qc2.shape)
    # # (17520, 3) (17520,)
    # print(np.all(nee_f==nee_f2[:,0]), np.all(nee_f==nee_f2[:,1]), np.all(nee_f==nee_f2[:,2]))
    # # True True True

    # # 2D err
    # nee_std  = gapfill(jdate, nee,  rg, tair, vpd, data_flag=(qcnee>1), undef=undef, err=True)
    # nee_std2 = gapfill(jdate, nee2, rg, tair, vpd, data_flag=(qcnee>1), undef=undef, err=True)
    # print(np.all(nee_std==nee_std2[:,0]), np.all(nee_std==nee_std2[:,1]), np.all(nee_std==nee_std2[:,2]))
    # # True True True
